import shlex
from typing import Any, Dict, Optional, Sequence, Union

from pydantic.v1 import utils as pydantic_v1_utils

import servicefoundry.lib.dao.version as version_lib
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.model.entity import Application, Deployment, TriggerJobResult
from servicefoundry.lib.util import (
    find_list_paths,
    get_application_fqn_from_deployment_fqn,
)
from servicefoundry.logger import logger


def list_applications(
    application_type: str,
    workspace_fqn: Optional[str] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    if workspace_fqn:
        workspace = client.get_id_from_fqn(fqn=workspace_fqn, fqn_type="workspace")
        applications = client.list_applications(workspace_id=workspace["workspaceId"])
    else:
        applications = client.list_applications()

    if application_type != "all":
        applications = [
            application
            for application in applications
            if application.deployment.manifest.type == application_type
        ]
    return applications


def get_application(
    application_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    application = client.get_application_info(
        application_id=application["applicationId"]
    )
    return application


def delete_application(
    application_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Dict[str, Any]:
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    response = client.remove_application(application_id=application["applicationId"])
    return response


def redeploy_application(
    application_fqn: str,
    version: int,
    wait: bool,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Deployment:
    from servicefoundry.v2.lib.deployable_patched_models import Application

    client = client or ServiceFoundryServiceClient()

    deployment_info = version_lib.get_version(
        application_fqn=application_fqn, version=version
    )

    manifest = deployment_info.manifest.dict()

    application_id = deployment_info.applicationId
    application_info = client.get_application_info(application_id=application_id)
    workspace_fqn = application_info.workspace.fqn

    application = Application.parse_obj(manifest)
    deployment = application.deploy(workspace_fqn=workspace_fqn, wait=wait)
    return deployment


def list_job_runs(
    application_fqn: str,
    max_results: Optional[int] = None,
    offset: Optional[int] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    response = client.list_job_runs(
        application_id=application["applicationId"], limit=max_results, offset=offset
    )
    return response


def get_job_run(
    application_fqn: str,
    job_run_name: str,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    application = client.get_id_from_fqn(fqn=application_fqn, fqn_type="app")
    response = client.get_job_run(
        application_id=application["applicationId"], job_run_name=job_run_name
    )
    return response


def trigger_job(
    application_fqn: str,
    component_name: str = None,
    command: Optional[Union[str, Sequence[str]]] = None,
    params: Optional[Dict[str, str]] = None,
) -> TriggerJobResult:
    """
    Trigger a Job on Truefoundry platform

    Args:
        application_fqn: Fully Qualified Name of the Deployed Job (without the version number)
        component_name: Name of the component to trigger the job on. Required in case of type `application`, defaults to `None`.
        command: command to run the job with, defaults to `None`. Can be a `str` or `List[str]`
            When `None`, the job is triggered with configured command at the time of deployment.
            When passed as a list, the command will be joined using `shlex.join`
        params: A dict mapping from parameter names (as defined in the job spec) to string values

    Returns:
        TriggerJobResult: metadata returning status of job trigger
    """
    if params and command:
        raise ValueError(
            "`command` and `params` arguments are mutually exclusive. Please pass only one of them"
        )

    try:
        # If user is passing in deployment fqn copied from UI, till we change the fqns on UI
        application_fqn = get_application_fqn_from_deployment_fqn(application_fqn)
        logger.warning(
            "Detected version number in `application_fqn`. "
            "Automatically discarding the version number. "
            f"This automatic conversion will be removed in future. "
            f"Please pass {application_fqn!r} as the value."
        )
    except ValueError:
        pass
    client = ServiceFoundryServiceClient()
    _application_info = client.get_application_info_by_fqn(
        application_fqn=application_fqn
    )
    application_info = client.get_application_info(
        application_id=_application_info.applicationId
    )
    command_str = ""
    message = ""
    if not params and not command:
        message = "Job triggered with pre-configured command"
    elif command:
        if not isinstance(command, str):
            command_str = shlex.join(command).strip()
        else:
            command_str = command.strip()
        message = f"Job triggered with command {command_str!r}"
    elif params:
        # Check if params has any non string values
        for key, value in params.items():
            if not isinstance(value, str):
                raise ValueError(
                    f"Invalid value {value!r} for key {key!r}. "
                    "Only string values are allowed for `params`"
                )
        command_str = ""
        message = f"Job triggered with params {params!r}"
    result = client.trigger_job(
        deployment_id=application_info.activeDeploymentId,
        component_name=component_name,
        command=command_str if command_str else None,
        params=params if params else None,
    )
    previous_runs_url = f"{client.base_url.strip('/')}/deployments/{application_info.id}?tab=previousRuns"
    logger.info(
        f"{message}.\n"
        f"You can check the status of your job run at {previous_runs_url}"
    )
    return result


def get_patched_application_definition(
    application: Application,
    manifest_patch: Dict[str, Any],
) -> Dict[str, Any]:
    # TODO: define ApplicationPatch type but for now
    # I am adding a manual check for name since name patch can
    # create a new application
    if (
        manifest_patch.get("name")
        and manifest_patch.get("name") != application.deployment.manifest.name
    ):
        raise Exception(
            f"Cannot change name of application from `{application.deployment.manifest.name}` to `{manifest_patch.get('name')}`"
        )

    patch_list_paths = find_list_paths(manifest_patch)

    for path in patch_list_paths:
        logger.warn(
            f"You are patching the value at {path}. Note that updating array-type objects will replace the entire object."
        )

    return pydantic_v1_utils.deep_update(
        application.deployment.manifest.dict(), manifest_patch
    )


def terminate_job_run(
    application_fqn: str,
    job_run_name: str,
):
    try:
        application_fqn = get_application_fqn_from_deployment_fqn(application_fqn)
        logger.warning(
            "Detected version number in `application_fqn`. "
            "Automatically discarding the version number. "
            f"This automatic conversion will be removed in future. "
            f"Please pass {application_fqn!r} as the value."
        )
    except ValueError:
        pass
    client = ServiceFoundryServiceClient()
    _application_info = client.get_application_info_by_fqn(
        application_fqn=application_fqn
    )
    application_info = client.get_application_info(
        application_id=_application_info.applicationId
    )

    response = client.terminate_job_run(
        deployment_id=application_info.activeDeploymentId,
        job_run_name=job_run_name,
    )
    return response



from typing import Any, Dict, List, Optional

from langchain.chat_models.base import SimpleChatModel
from langchain.pydantic_v1 import Extra, Field, root_validator
from langchain.schema.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    HumanMessage,
    SystemMessage,
)

from servicefoundry.langchain.utils import (
    ModelParameters,
    requests_retry_session,
    validate_tfy_environment,
)
from servicefoundry.logger import logger


class TrueFoundryChat(SimpleChatModel):
    """`TrueFoundry LLM Gateway` chat models API.

    To use, you must have the environment variable ``TFY_API_KEY`` set with your API key and ``TFY_HOST`` set with your host or pass it as a named parameter to the constructor.
    """

    model: str = Field(description="The model to use for chat.")
    """The model to use for chat."""
    tfy_llm_gateway_url: Optional[str] = Field(default=None)
    """TrueFoundry LLM Gateway endpoint URL. Automatically inferred from env var `TFY_LLM_GATEWAY_URL` if not provided."""
    tfy_api_key: Optional[str] = Field(default=None)
    """TrueFoundry API Key. Automatically inferred from env var `TFY_API_KEY` if not provided."""
    model_parameters: Optional[dict] = Field(default_factory=dict)
    """Model parameters"""
    request_timeout: int = Field(default=30)
    """The timeout for the request in seconds."""
    max_retries: int = Field(default=5)
    """The number of retries for HTTP requests."""
    retry_backoff_factor: float = Field(default=0.3)
    """The backoff factor for exponential backoff during retries."""
    system_prompt: str = Field(default="You are a AI assistant")

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        allow_population_by_field_name = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        values = validate_tfy_environment(values)
        if not values["tfy_api_key"]:
            raise ValueError(
                f"Did not find `tfy_api_key`, please add an environment variable"
                f" `TFY_API_KEY` which contains it, or pass"
                f"  `tfy_api_key` as a named parameter."
            )
        if not values["tfy_llm_gateway_url"]:
            raise ValueError(
                f"Did not find `tfy_llm_gateway_url`, please add an environment variable"
                f" `TFY_LLM_GATEWAY_URL` which contains it, or pass"
                f"  `tfy_llm_gateway_url` as a named parameter."
            )
        return values

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "truefoundry-chat"

    def _call(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> str:

        if len(messages) == 0:
            raise ValueError("No messages provided to chat.")

        if not isinstance(messages[0], SystemMessage):
            messages.insert(0, SystemMessage(content=self.system_prompt))

        message_dicts = [
            TrueFoundryChat._convert_message_to_dict(message) for message in messages
        ]

        params = {**self.model_parameters} if self.model_parameters else {}
        if stop:
            params["stop_sequences"] = stop

        payload = {
            "messages": message_dicts,
            "model": {
                "name": self.model,
                "parameters": params,
            },
        }
        session = requests_retry_session(
            retries=self.max_retries, backoff_factor=self.retry_backoff_factor
        )

        url = f"{self.tfy_llm_gateway_url}/api/inference/chat"
        logger.debug(f"Chat using - model: {self.model} at endpoint: {url}")
        response = session.post(
            url=url,
            json=payload,
            headers={
                "Authorization": f"Bearer {self.tfy_api_key}",
            },
            timeout=self.request_timeout,
        )
        response.raise_for_status()
        output = response.json()
        return output["text"]

    @staticmethod
    def _convert_message_to_dict(message: BaseMessage) -> dict:
        if isinstance(message, ChatMessage):
            message_dict = {"role": message.role, "content": message.content}
        elif isinstance(message, HumanMessage):
            message_dict = {"role": "user", "content": message.content}
        elif isinstance(message, AIMessage):
            message_dict = {"role": "assistant", "content": message.content}
        elif isinstance(message, SystemMessage):
            message_dict = {"role": "system", "content": message.content}
        else:
            raise ValueError(f"Got unknown message type: {message}")
        if message.additional_kwargs:
            logger.debug(
                "Additional message arguments are unsupported by TrueFoundry LLM Gateway "
                " and will be ignored: %s",
                message.additional_kwargs,
            )
        return message_dict



import json

import rich_click as click
import yaml

from servicefoundry.cli.const import GROUP_CLS
from servicefoundry.lib.dao import application as application_lib


@click.group(
    name="deploy",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Deploy application to Truefoundry",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    default="./servicefoundry.yaml",
    help="Path to servicefoundry.yaml file",
    show_default=True,
)
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    required=True,
    help="FQN of the Workspace to deploy to",
)
@click.option(
    "--wait/--no-wait",
    "--wait/--no_wait",
    is_flag=True,
    show_default=True,
    default=True,
    help="Wait and tail the deployment progress",
)
def deploy_v2_command(file: str, workspace_fqn: str, wait: bool):
    from servicefoundry.v2.lib.deployable_patched_models import Application

    with open(file, "r") as f:
        application_definition = yaml.safe_load(f)

    application = Application.parse_obj(application_definition)
    _deployment = application.deploy(workspace_fqn=workspace_fqn, wait=wait)
    # print_json(_deployment)


@click.group(
    name="patch-application",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Deploy application with patches to Truefoundry",
)
@click.option(
    "-f",
    "--patch-file",
    "--patch_file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to yaml patch file",
    show_default=True,
)
@click.option(
    "-p",
    "--patch",
    type=click.STRING,
    help="Patch in JSON format provided as a string.",
    show_default=True,
)
@click.option(
    "-a",
    "--application_fqn",
    "--application-fqn",
    type=click.STRING,
    required=True,
    help="FQN of the Application to patch and deploy",
)
@click.option(
    "--wait/--no-wait",
    "--wait/--no_wait",
    is_flag=True,
    show_default=True,
    default=True,
    help="Wait and tail the deployment progress",
)
def deploy_patch_v2_command(
    patch_file: str, application_fqn: str, patch: str, wait: bool
):
    from servicefoundry.v2.lib.deployable_patched_models import Application

    manifest_patch_obj = None
    if not patch_file and not patch:
        raise Exception("You need to either provide --file or --patch.")
    elif patch and patch_file:
        raise Exception("You can only provide one of --file and --patch")
    elif patch:
        try:
            manifest_patch_obj = json.loads(patch)
        except json.decoder.JSONDecodeError:
            raise Exception("Invalid JSON provided as --patch")
    elif patch_file:
        with open(patch_file, "r") as f:
            manifest_patch_obj = yaml.safe_load(f)

    if not manifest_patch_obj or type(manifest_patch_obj) != dict:
        raise Exception("Invalid patch, aborting deployment.")

    tfy_application = application_lib.get_application(application_fqn=application_fqn)
    patched_application_obj = application_lib.get_patched_application_definition(
        application=tfy_application, manifest_patch=manifest_patch_obj
    )

    application = Application.parse_obj(patched_application_obj)
    application.deploy(workspace_fqn=tfy_application.workspace.fqn, wait=wait)



import enum
import warnings
from typing import Any, Dict, List, Mapping, Optional
from urllib.parse import urljoin

import requests
from requests.auth import HTTPBasicAuth

from servicefoundry.lib.auth.servicefoundry_session import ServiceFoundrySession
from servicefoundry.pydantic_v1 import root_validator

try:
    from langchain.callbacks.manager import CallbackManagerForLLMRun
    from langchain.llms.base import LLM
    from langchain.llms.utils import enforce_stop_tokens
except Exception as ex:
    raise Exception(
        "Failed to import langchain."
        " Please install langchain by using `pip install langchain` command"
    ) from ex


class _ModelServerImpl(str, enum.Enum):
    MLSERVER = "MLSERVER"
    TGI = "TGI"
    VLLM = "VLLM"


def _get_model_server_and_validate_if_mlserver(endpoint_url, auth, model_name=None):
    try:
        response = requests.get(urljoin(endpoint_url, "info"), json={}, auth=auth)
        if response.status_code == 200:
            return _ModelServerImpl.TGI, None
        elif response.status_code == 404:
            # We are not using TGI, try for mlserver
            response = requests.post(
                urljoin(endpoint_url, "v2/repository/index"), json={}, auth=auth
            )
            if response.status_code == 200:
                models = response.json()
                if len(models) == 0:
                    raise ValueError("No model is deployed in the model server")
                model_names = [m.get("name") for m in models]
                if model_name and model_name not in model_names:
                    raise ValueError(
                        f"Model {model_name!r} is not available in the model server. "
                        f"Available models {model_names!r}"
                    )
                if not model_name and len(model_names) > 1:
                    raise ValueError(
                        f"Please pass `model_name` while instantiating `TruefoundryLLM`. "
                        f"Available models are {model_names!r} "
                    )
                if model_name:
                    return _ModelServerImpl.MLSERVER, model_name
                return _ModelServerImpl.MLSERVER, model_names[0]
            if response.status_code == 404:
                return _ModelServerImpl.VLLM, None
        response.raise_for_status()
    except Exception as e:
        raise Exception(f"Error raised by inference API: {e}") from e


# TODO (chiragjn): Refactor this into separate implementations for each model server


class TruefoundryLLM(LLM):
    """Wrapper around TFY model deployment.
    To use this class, you need to have the langchain library installed.
    Example:
        .. code-block:: python
            from servicefoundry.langchain import TruefoundryLLM
            endpoint_url = (
                "https://pythia-70m-model-model-catalogue.demo2.truefoundry.tech"
            )
            model = TruefoundryLLM(
                endpoint_url=endpoint_url,
                parameters={
                    "max_new_tokens": 100,
                    "temperature": 0.7,
                    "top_k": 5,
                    "top_p": 0.9
                }
            )
    """

    endpoint_url: str
    model_name: Optional[str] = None
    auth: Optional[HTTPBasicAuth] = None
    parameters: Optional[Dict[str, Any]] = None
    model_server_impl: Optional[_ModelServerImpl] = None

    @root_validator(pre=False)
    def validate_model_server_and_name(cls, values: Dict):
        warnings.warn(
            message=f"{cls.__name__} is deprecated and will be removed soon. Please use `TrueFoundryLLM` or `TrueFoundryChat` to invoke models using the new TrueFoundry LLM Gateway",
            category=DeprecationWarning,
            stacklevel=2,
        )
        endpoint_url = values["endpoint_url"]
        model_name = values.get("model_name")
        auth = values.get("auth")
        model_server_impl, model_name = _get_model_server_and_validate_if_mlserver(
            endpoint_url=endpoint_url, model_name=model_name, auth=auth
        )
        values["model_server_impl"] = model_server_impl
        if model_server_impl == _ModelServerImpl.MLSERVER:
            values["model_name"] = model_name
        return values

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {
            "endpoint_url": self.endpoint_url,
            "model_name": self.model_name,
        }

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "tfy_model_deployment"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **params: Any,
    ) -> str:
        """Call out to the deployed model
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
        Returns:
            The string generated by the model.
        Example:
            .. code-block:: python
                response = model("Tell me a joke.")
        """
        _params_already_set = self.parameters or {}
        params = {**_params_already_set, **params, "return_full_text": False}

        if self.model_server_impl == _ModelServerImpl.MLSERVER:
            generate_path = f"v2/models/{self.model_name}/infer/simple"
            payload = {"inputs": prompt, "parameters": params}
        elif self.model_server_impl == _ModelServerImpl.TGI:
            generate_path = "generate"
            payload = {"inputs": prompt, "parameters": params}
        elif self.model_server_impl == _ModelServerImpl.VLLM:
            generate_path = "generate"
            payload = {**params, "prompt": prompt}
        else:
            raise ValueError(f"No known generate path for {self.model_server_impl}")
        url = urljoin(self.endpoint_url, generate_path)

        try:
            response = requests.post(url, json=payload, auth=self.auth)
            response.raise_for_status()
        except Exception as e:
            raise Exception(f"Error raised by inference API: {e}") from e
        response_dict = response.json()
        if "error" in response_dict:
            raise ValueError(
                f"Error raised by inference API: {response_dict['error']!r}"
            )

        if self.model_server_impl == _ModelServerImpl.MLSERVER:
            inference_result = response_dict[0]
        elif self.model_server_impl == _ModelServerImpl.TGI:
            inference_result = response_dict
        elif self.model_server_impl == _ModelServerImpl.VLLM:
            inference_result = response_dict
        else:
            raise ValueError(
                f"Unknown model server {self.model_server_impl}, cannot parse response"
            )

        if "generated_text" in inference_result:
            text = inference_result["generated_text"]
        elif "summarization" in inference_result:
            text = inference_result["summary_text"]
        elif "text" in inference_result:
            text = inference_result["text"]
        else:
            raise ValueError(f"Could not parse inference response: {response_dict!r}")

        if isinstance(text, list):
            text = text[0]

        if stop:
            text = enforce_stop_tokens(text, stop)

        return text


class TruefoundryPlaygroundLLM(LLM):
    """Wrapper around TFY Playground.
    To use this class, you need to have the langchain library installed.
    Example:
        .. code-block:: python
            from servicefoundry.langchain import TruefoundryPlaygroundLLM
            import os
            # Note: Login using servicefoundry login --host <https://example-domain.com>
            model = TruefoundryPlaygroundLLM(
                model_name="vicuna-13b",
                parameters={
                    "maximumLength": 100,
                    "temperature": 0.7,
                    "topP": 0.9,
                    "repetitionPenalty": 1
                }
            )
            response = model.predict("Enter the prompt here")
    """

    model_name: str
    parameters: Optional[Dict[str, Any]] = None
    provider: str = "truefoundry-public"

    @root_validator(pre=False)
    def validate_model_server_and_name(cls, values: Dict):
        warnings.warn(
            message=f"{cls.__name__} is deprecated and will be removed soon. Please use `TrueFoundryLLM` or `TrueFoundryChat` to invoke models using the new TrueFoundry LLM Gateway",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return values

    @property
    def _get_model(self) -> str:
        """returns the model name"""
        return self.model_name

    @property
    def _get_provider(self) -> str:
        """Returns the provider name"""
        return self.provider

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "tfy_playground"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        **params: Any,
    ) -> str:
        """Call out to the deployed model
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
        Returns:
            The string generated by the model.
        Example:
            .. code-block:: python
                response = model("I have a joke for you...")
        """
        _params_already_set = self.parameters or {}
        params = {**_params_already_set, **params}
        if stop:
            params["stopSequences"] = stop
        session = ServiceFoundrySession()

        if not session:
            raise Exception(
                f"Unauthenticated: Please login using servicefoundry login --host <https://example-domain.com>"
            )

        host = session.base_url

        if host[-1] == "/":
            host = host[: len(host) - 1]

        url = f"{host}/llm-playground/api/inference/text"
        headers = {"Authorization": f"Bearer {session.access_token}"}

        json = {
            "prompt": prompt,
            "models": [
                {
                    "name": self.model_name,
                    "provider": self.provider,
                    "tag": self.model_name,
                    "parameters": params,
                }
            ],
        }

        try:
            response = requests.post(url=url, headers=headers, json=json)
            response.raise_for_status()
        except Exception as ex:
            raise Exception(f"Error inferencing the model: {ex}") from ex

        data = response.json()
        text = data[0].get("text")
        if stop:
            text = enforce_stop_tokens(text, stop)
        return text



import os
from typing import Dict, List, Optional
from urllib.parse import urljoin

import requests
from langchain.pydantic_v1 import BaseModel
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from servicefoundry.lib.auth.servicefoundry_session import ServiceFoundrySession


class ModelParameters(BaseModel):
    temperature: Optional[float]
    maximum_length: Optional[int]
    top_p: Optional[float]
    top_k: Optional[int]
    repetition_penalty: Optional[float]
    frequency_penalty: Optional[float]
    presence_penalty: Optional[float]
    stop_sequences: Optional[List[str]]


def validate_tfy_environment(values: Dict):
    gateway_url = values["tfy_llm_gateway_url"] or os.getenv("TFY_LLM_GATEWAY_URL")
    api_key = values["tfy_api_key"] or os.getenv("TFY_API_KEY")

    if gateway_url and api_key:
        values["tfy_llm_gateway_url"] = gateway_url
        values["tfy_api_key"] = api_key
        return values

    sfy_session = ServiceFoundrySession()
    if not sfy_session:
        raise Exception(
            f"Unauthenticated: Please login using servicefoundry login --host <https://example-domain.com>"
        )

    if not gateway_url:
        gateway_url = urljoin(sfy_session.base_url, "/api/llm")

    if not api_key:
        api_key = sfy_session.access_token

    values["tfy_llm_gateway_url"] = gateway_url
    values["tfy_api_key"] = api_key
    return values


def requests_retry_session(
    retries=5,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 503, 504),
    method_whitelist=frozenset({"GET", "POST"}),
    session=None,
):
    """
    Returns a `requests` session with retry capabilities for certain HTTP status codes.

    Args:
        retries (int): The number of retries for HTTP requests.
        backoff_factor (float): The backoff factor for exponential backoff during retries.
        status_forcelist (tuple): A tuple of HTTP status codes that should trigger a retry.
        method_whitelist (frozenset): The set of HTTP methods that should be retried.
        session (requests.Session, optional): An optional existing requests session to use.

    Returns:
        requests.Session: A session with retry capabilities.
    """
    # Implementation taken from https://www.peterbe.com/plog/best-practice-with-retries-with-requests
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        status=retries,
        backoff_factor=backoff_factor,
        allowed_methods=method_whitelist,
        status_forcelist=status_forcelist,
        respect_retry_after_header=True,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



import datetime
import functools
import json

from rich import box
from rich import print_json as _rich_print_json
from rich.table import Table

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.console import console
from servicefoundry.cli.const import DISPLAY_DATETIME_FORMAT
from servicefoundry.json_util import json_default_encoder


def print_json(data, default=json_default_encoder):
    return _rich_print_json(
        json.dumps(data, default=default), highlight=False, default=default
    )


NO_WRAP_COLUMNS = {"fqn"}


def get_table(title):
    return Table(title=title, show_lines=False, safe_box=True, box=box.MINIMAL)


def stringify(x):
    if isinstance(x, datetime.datetime):
        return x.astimezone().strftime(DISPLAY_DATETIME_FORMAT)
    elif isinstance(x, str):
        return x
    else:
        return str(x)


def display_time_passed(seconds: int):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)
    result = ""
    if d != 0:
        result = f"{int(d)}D"
    elif h != 0 and m != 0:
        result = f"{int(h)}h {int(m)}m"
    elif h != 0:
        result = f"{int(h)}h"
    elif m != 0 and s != 0:
        result = f"{int(m)}m {int(s)}s"
    elif m != 0:
        result = f"{int(m)}m"
    elif s != 0:
        result = f"{int(s)}s"
    return result


def print_entity_list(title, items):
    items = [item.list_row_data() for item in items]
    if CliConfig.get("json"):
        print_json(data=items)
        return

    table = get_table(title)

    columns = []
    if items:
        columns = items[0].keys()
        for column in columns:
            no_wrap = column in NO_WRAP_COLUMNS
            table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)

    for item in items:
        row = []
        for c in columns:
            row.append(stringify(item[c]))
        table.add_row(*row)
    console.print(table)


def print_obj(title, item, columns=None):
    if CliConfig.get("json"):
        print_json(data=item)
        return

    table = get_table(title)

    if not columns:
        columns = item.keys()

    # transpose
    keys, columns = columns, ["key", "value"]

    for column in columns:
        no_wrap = column in NO_WRAP_COLUMNS
        table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)
    for key in keys:
        table.add_row(f"[bold]{stringify(key)}[/]", stringify(item[key]))
    console.print(table)


def print_entity_obj(title, entity):
    if CliConfig.get("json"):
        print_json(data=entity)
        return

    table = get_table(title)

    columns = entity.get_data().keys()

    # transpose
    keys, columns = columns, ["key", "value"]

    for column in columns:
        no_wrap = "FQN" in column or "Name" in column
        table.add_column(column, justify="left", overflow="fold", no_wrap=no_wrap)
    entity_data = entity.get_data()
    for key in keys:
        table.add_row(f"[bold]{stringify(key)}[/]", stringify(entity_data[key]))
    console.print(table)



from typing import Union

from servicefoundry.auto_gen import models
from servicefoundry.lib.model.entity import Deployment
from servicefoundry.pydantic_v1 import BaseModel, Field, conint, constr
from servicefoundry.v2.lib.deploy import deploy_component


class DeployablePatchedModelBase(BaseModel):
    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Service(models.Service, DeployablePatchedModelBase):
    type: constr(regex=r"service") = "service"
    resources: models.Resources = Field(default_factory=models.Resources)
    # This is being patched because cue export marks this as a "number"
    replicas: Union[conint(ge=0, le=100), models.ServiceAutoscaling] = Field(
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone\n+sort=3",
    )


class Job(models.Job, DeployablePatchedModelBase):
    type: constr(regex=r"job") = "job"
    resources: models.Resources = Field(default_factory=models.Resources)


class Notebook(models.Notebook, DeployablePatchedModelBase):
    type: constr(regex=r"^notebook$") = "notebook"
    resources: models.Resources = Field(default_factory=models.Resources)


class Codeserver(models.Codeserver, DeployablePatchedModelBase):
    type: constr(regex=r"^codeserver$") = "codeserver"


class Helm(models.Helm, DeployablePatchedModelBase):
    type: constr(regex=r"^helm$") = "helm"


class Volume(models.Volume, DeployablePatchedModelBase):
    type: constr(regex=r"^volume$") = "volume"


class AsyncService(models.AsyncService, DeployablePatchedModelBase):
    type: constr(regex=r"^async-service$") = "async-service"
    replicas: Union[conint(ge=0, le=100), models.AsyncServiceAutoscaling] = 1
    resources: models.Resources = Field(default_factory=models.Resources)


class SSHServer(models.SSHServer, DeployablePatchedModelBase):
    type: constr(regex=r"^ssh-server$") = "ssh-server"


class Application(models.Application, DeployablePatchedModelBase):
    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(
            component=self.__root__, workspace_fqn=workspace_fqn, wait=wait
        )



import os
import re
from typing import Union

from servicefoundry.lib.const import (
    SFY_DEBUG_ENV_KEY,
    SFY_EXPERIMENTAL_ENV_KEY,
    SFY_INTERNAL_ENV_KEY,
)


def is_debug_env_set() -> bool:
    return True if os.getenv(SFY_DEBUG_ENV_KEY) else False


def is_experimental_env_set() -> bool:
    # TODO (chiragjn): one of these need to be removed
    return (
        True
        if os.getenv(SFY_EXPERIMENTAL_ENV_KEY) or os.getenv(SFY_INTERNAL_ENV_KEY)
        else False
    )


def get_application_fqn_from_deployment_fqn(deployment_fqn: str) -> str:
    if not re.search(r":\d+$", deployment_fqn):
        raise ValueError(
            "Invalid `deployment_fqn` format. A deployment fqn is supposed to end with a version number"
        )
    application_fqn, _ = deployment_fqn.rsplit(":", 1)
    return application_fqn


def get_deployment_fqn_from_application_fqn(
    application_fqn: str, version: Union[str, int]
) -> str:
    return f"{application_fqn}:{version}"


def is_notebook():
    # https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    try:
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__
        if shell.__name__ == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell.__name__ == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        elif "google.colab" in str(shell):
            return True  # google colab notebook
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def find_list_paths(data, parent_key="", sep="."):
    list_paths = []
    if isinstance(data, dict):
        for key, value in data.items():
            new_key = f"{parent_key}{sep}{key}" if parent_key else key
            list_paths.extend(find_list_paths(value, new_key, sep))
    elif isinstance(data, list):
        list_paths.append(parent_key)
        for i, value in enumerate(data):
            new_key = f"{parent_key}[{i}]"
            list_paths.extend(find_list_paths(value, new_key, sep))
    return list_paths



from __future__ import annotations

import functools
import json
import os
import time
from datetime import datetime, timezone
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from urllib.parse import urljoin, urlparse

import requests
import socketio
from dateutil.tz import tzlocal
from packaging import version
from rich.status import Status
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from servicefoundry.io.output_callback import OutputCallBack
from servicefoundry.lib.auth.servicefoundry_session import ServiceFoundrySession
from servicefoundry.lib.clients.utils import request_handling
from servicefoundry.lib.const import API_SERVER_RELATIVE_PATH, VERSION_PREFIX
from servicefoundry.lib.model.entity import (
    Application,
    CreateDockerRepositoryResponse,
    Deployment,
    DockerRegistryCredentials,
    JobRun,
    TenantInfo,
    Token,
    TriggerJobResult,
    Workspace,
    WorkspaceResources,
)
from servicefoundry.lib.win32 import allow_interrupt
from servicefoundry.logger import logger
from servicefoundry.pydantic_v1 import parse_obj_as
from servicefoundry.v2.lib.models import (
    AppDeploymentStatusResponse,
    ApplicationFqnResponse,
    BuildResponse,
    DeploymentFqnResponse,
)
from servicefoundry.version import __version__

DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE = "DEPLOYMENT_LOGS"
BUILD_LOGS_SUBSCRIBE_MESSAGE = "logs"

if TYPE_CHECKING:
    from servicefoundry.auto_gen.models import Application


def _upload_packaged_code(metadata, package_file):
    file_size = os.stat(package_file).st_size
    with open(package_file, "rb") as file_to_upload:
        with tqdm(
            total=file_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc="Uploading package",
        ) as progress_bar:
            wrapped_file = CallbackIOWrapper(
                progress_bar.update, file_to_upload, "read"
            )
            headers = metadata.get("headers", {})
            http_response = requests.put(
                metadata["url"], data=wrapped_file, headers=headers
            )

            if http_response.status_code not in [204, 201, 200]:
                raise RuntimeError(f"Failed to upload code {http_response.content}")


def check_min_cli_version(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        if __version__ != "0.0.0":
            # "0.0.0" indicates dev version
            client: "ServiceFoundryServiceClient" = args[0]
            # noinspection PyProtectedMember
            min_cli_version = client._get_min_cli_version_requirement()
            if version.parse(__version__) < version.parse(min_cli_version):
                raise Exception(
                    "You are using an outdated version of `servicefoundry`.\n"
                    f"Run `pip install servicefoundry>={min_cli_version}` to install the supported version.",
                )
        else:
            logger.debug("Ignoring minimum cli version check")

        return fn(*args, **kwargs)

    return inner


class ServiceFoundryServiceClient:
    def __init__(self, init_session: bool = True, base_url: Optional[str] = None):
        self._session: Optional[ServiceFoundrySession] = None
        if init_session:
            if base_url:
                logger.warning("Passed base url %r will be ignored", base_url)
            self._session = ServiceFoundrySession()
            base_url = self._session.base_url
        elif not base_url:
            raise Exception("Neither session, not base_url provided")

        self._base_url = base_url.strip("/")
        self._api_server_url = f"{self._base_url}/{API_SERVER_RELATIVE_PATH}"

    @property
    def base_url(self) -> str:
        return self._base_url

    def get_tenant_info(self) -> TenantInfo:
        res = requests.get(
            url=f"{self._api_server_url}/v1/tenant-id",
            params={"hostName": urlparse(self._api_server_url).netloc},
        )
        res = request_handling(res)
        return TenantInfo.parse_obj(res)

    @lru_cache(maxsize=3)
    def _get_min_cli_version_requirement(self) -> str:
        url = f"{self._api_server_url}/v1/min-cli-version"
        res = requests.get(url)
        res = request_handling(res)
        return res["minVersion"]

    def _get_header(self):
        if not self._session:
            return {}
        return {"Authorization": f"Bearer {self._session.access_token}"}

    @check_min_cli_version
    def get_id_from_fqn(self, fqn_type: str, fqn: str):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/{fqn_type}"
        res = requests.get(url, headers=self._get_header(), params={"fqn": fqn})
        return request_handling(res)

    @check_min_cli_version
    def list_workspace(self):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    @check_min_cli_version
    def list_workspaces(
        self,
        cluster_id: Optional[str] = None,
        workspace_name: Optional[str] = None,
        workspace_fqn: Optional[str] = None,
    ) -> List[Workspace]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        params = {}
        if cluster_id:
            params["clusterId"] = cluster_id
        if workspace_name:
            params["workspaceName"] = workspace_name
        if workspace_fqn:
            params["workspaceFqn"] = workspace_fqn
        res = requests.get(url, params=params, headers=self._get_header())
        response = request_handling(res)
        return parse_obj_as(List[Workspace], response)

    @check_min_cli_version
    def create_workspace(
        self,
        workspace_name: str,
        cluster_name: str,
        resources: WorkspaceResources,
    ) -> Workspace:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.post(
            url,
            json={
                "manifest": {
                    "cluster": cluster_name,
                    "name": workspace_name,
                    "resources": resources.dict(exclude_none=True),
                }
            },
            headers=self._get_header(),
        )
        res = request_handling(res)
        return Workspace.parse_obj(res)

    @check_min_cli_version
    def remove_workspace(self, workspace_id, force=False) -> Workspace:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace/{workspace_id}"
        force = json.dumps(
            force
        )  # this dumb conversion is required because `params` just casts as str
        res = requests.delete(url, headers=self._get_header(), params={"force": force})
        response = request_handling(res)
        return Workspace.parse_obj(response["workspace"])

    @check_min_cli_version
    def get_workspace_by_name(self, workspace_name, cluster_id):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={"name": workspace_name, "clusterId": cluster_id},
        )
        return request_handling(res)

    @check_min_cli_version
    def get_workspace(self, workspace_id):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace/{workspace_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    @check_min_cli_version
    def get_workspace_by_fqn(self, workspace_fqn: str) -> List[Workspace]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/workspace"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={"fqn": workspace_fqn},
        )
        response = request_handling(res)
        return parse_obj_as(List[Workspace], response)

    @check_min_cli_version
    def list_deployments(self, workspace_id: str = None):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment"
        params = {}
        if workspace_id:
            params["workspaceId"] = workspace_id
        res = requests.get(url=url, params=params, headers=self._get_header())
        return request_handling(res)

    @check_min_cli_version
    def list_cluster(self):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/cluster"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    @check_min_cli_version
    def get_cluster(self, cluster_id):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/cluster/{cluster_id}"
        res = requests.get(url, headers=self._get_header())
        return request_handling(res)

    @check_min_cli_version
    def get_presigned_url(self, space_name, service_name, env):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment/code-upload-url"
        res = requests.post(
            url,
            json={
                "workspaceFqn": space_name,
                "serviceName": service_name,
                "stage": env,
            },
            headers=self._get_header(),
        )
        return request_handling(res)

    @check_min_cli_version
    def upload_code_package(
        self, workspace_fqn: str, component_name: str, package_local_path: str
    ) -> str:
        http_response = self.get_presigned_url(
            space_name=workspace_fqn, service_name=component_name, env="default"
        )
        _upload_packaged_code(metadata=http_response, package_file=package_local_path)

        return http_response["uri"]

    @check_min_cli_version
    def deploy_application(
        self, workspace_id: str, application: Application
    ) -> Deployment:
        data = {
            "workspaceId": workspace_id,
            "name": application.name,
            "manifest": application.dict(exclude_none=True),
        }
        logger.debug(json.dumps(data))
        url = f"{self._api_server_url}/{VERSION_PREFIX}/deployment"
        deploy_response = requests.post(url, json=data, headers=self._get_header())
        response = request_handling(deploy_response)
        return Deployment.parse_obj(response["deployment"])

    def _get_log_print_line(self, log: dict):
        timestamp = int(log["time"]) / 1e6

        time_obj = datetime.fromtimestamp(timestamp / 1000.0, tz=timezone.utc)
        time_obj.replace(tzinfo=timezone.utc)
        local_time = time_obj.astimezone(tzlocal())
        local_time_str = local_time.isoformat()
        return f'[{local_time_str}] {log["log"].strip()}'

    def _tail_logs(
        self,
        tail_logs_url: str,
        query_dict: dict,
        # NOTE: Rather making this printer callback an argument,
        # we should have global printer callback
        # which will be initialized based on the running env (cli, lib, notebook)
        subscribe_message: str,
        socketio_path: str = "socket.io",
        callback=OutputCallBack(),
        wait=True,
    ):

        sio = socketio.Client(request_timeout=60)
        callback.print_line("Waiting for the task to start...")

        @sio.on(subscribe_message)
        def logs(data):
            try:
                _log = json.loads(data)
                callback.print_line(self._get_log_print_line(_log["body"]))
            except Exception:
                logger.exception(f"Error while parsing log line, {data!r}")

        def sio_disconnect_no_exception():
            try:
                sio.disconnect()
            except Exception:
                logger.exception("Error while disconnecting from socket connection")

        with allow_interrupt(sio_disconnect_no_exception):
            sio.connect(
                tail_logs_url,
                transports="websocket",
                headers=self._get_header(),
                socketio_path=socketio_path,
            )
            # TODO: We should have have a timeout here. `emit` does
            #   not support timeout. Explore `sio.call`.
            sio.emit(
                subscribe_message,
                json.dumps(query_dict),
            )
            if wait:
                sio.wait()

    @check_min_cli_version
    def get_deployment(self, application_id: str, deployment_id: str) -> Deployment:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments/{deployment_id}"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return Deployment.parse_obj(res)

    @check_min_cli_version
    def get_deployment_statuses(
        self, application_id: str, deployment_id: str
    ) -> List[AppDeploymentStatusResponse]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments/{deployment_id}/statuses"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return parse_obj_as(List[AppDeploymentStatusResponse], res)

    @check_min_cli_version
    def get_deployment_build_response(
        self, application_id: str, deployment_id: str
    ) -> List[BuildResponse]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments/{deployment_id}/builds"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return parse_obj_as(List[BuildResponse], res)

    def _get_deployment_logs(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        job_run_name: Optional[str] = None,
        start_ts_nano: Optional[int] = None,
        end_ts_nano: Optional[int] = None,
        limit: Optional[int] = None,
        num_logs_to_ignore: Optional[int] = None,
    ) -> List:
        get_logs_query = {"applicationId": application_id}
        if deployment_id:
            get_logs_query["deploymentId"] = deployment_id
        data = {"getLogsQuery": json.dumps(get_logs_query)}
        if start_ts_nano:
            data["startTs"] = str(start_ts_nano)
        if end_ts_nano:
            data["endTs"] = str(end_ts_nano)
        if limit:
            data["limit"] = str(limit)
        if num_logs_to_ignore:
            data["numLogsToIgnore"] = int(num_logs_to_ignore)
        if job_run_name:
            data["jobRunName"] = job_run_name

        url = f"{self._api_server_url}/{VERSION_PREFIX}/logs/{workspace_id}"
        res = requests.get(url=url, params=data, headers=self._get_header())
        res = request_handling(res)
        return list(res["logs"])

    @check_min_cli_version
    def tail_build_logs(
        self,
        build_response: BuildResponse,
        callback=OutputCallBack(),
        wait: bool = True,
    ):
        tail_logs_obj = json.loads(build_response.tailLogsUrl)
        self._tail_logs(
            tail_logs_url=tail_logs_obj["uri"],
            socketio_path=tail_logs_obj["path"],
            query_dict={
                "pipelineRunName": build_response.name,
                "startTs": build_response.logsStartTs,
            },
            callback=callback,
            wait=wait,
            subscribe_message=BUILD_LOGS_SUBSCRIBE_MESSAGE,
        )

    @check_min_cli_version
    def tail_logs_for_deployment(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        start_ts: int,
        limit: int,
        callback=OutputCallBack(),
        wait: bool = True,
    ):
        self._tail_logs(
            tail_logs_url=urljoin(
                self._api_server_url, f"/?type={DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE}"
            ),
            query_dict={
                "workspaceId": workspace_id,
                "startTs": str(int(start_ts * 1e6)),
                "limit": limit,
                "getLogsQuery": {
                    "applicationId": application_id,
                    "deploymentId": deployment_id,
                },
            },
            callback=callback,
            wait=wait,
            subscribe_message=DEPLOYMENT_LOGS_SUBSCRIBE_MESSAGE,
        )

    @check_min_cli_version
    def poll_logs_for_deployment(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        job_run_name: Optional[str],
        start_ts: int,
        limit: int,
        poll_interval_seconds: int,
        callback=OutputCallBack(),
    ):
        start_ts_nano = int(start_ts * 1e6)

        with Status(status="Polling for logs") as spinner:
            num_logs_to_ignore = 0

            while True:
                logs_list = self._get_deployment_logs(
                    workspace_id=workspace_id,
                    application_id=application_id,
                    deployment_id=deployment_id,
                    job_run_name=job_run_name,
                    start_ts_nano=start_ts_nano,
                    limit=limit,
                    num_logs_to_ignore=num_logs_to_ignore,
                )

                if len(logs_list) == 0:
                    logger.warning("Did not receive any logs")
                    time.sleep(poll_interval_seconds)
                    continue

                for log in logs_list:
                    callback.print_line(self._get_log_print_line(log))

                last_log_time = logs_list[-1]["time"]
                num_logs_to_ignore = 0
                for log in reversed(logs_list):
                    if log["time"] != last_log_time:
                        break
                    num_logs_to_ignore += 1

                start_ts_nano = int(last_log_time)
                spinner.update(status=f"Waiting for {poll_interval_seconds} secs.")
                time.sleep(poll_interval_seconds)

    @check_min_cli_version
    def fetch_deployment_logs(
        self,
        workspace_id: str,
        application_id: str,
        deployment_id: str,
        job_run_name: Optional[str],
        start_ts: Optional[int],
        end_ts: Optional[int],
        limit: Optional[int],
        callback=OutputCallBack(),
    ):
        logs_list = self._get_deployment_logs(
            workspace_id=workspace_id,
            application_id=application_id,
            deployment_id=deployment_id,
            job_run_name=job_run_name,
            start_ts_nano=int(start_ts * 1e6),
            end_ts_nano=int(end_ts * 1e6),
            limit=limit,
        )
        for log in logs_list:
            callback.print_line(self._get_log_print_line(log))

    @check_min_cli_version
    def fetch_build_logs(
        self,
        build_response: BuildResponse,
        callback=OutputCallBack(),
    ) -> None:
        url = build_response.getLogsUrl
        res = requests.get(url=url, headers=self._get_header())
        logs_list = request_handling(res)
        for log in logs_list["logs"]:
            # TODO: Have to establish a log line format that includes timestamp, level, message
            callback.print_line(self._get_log_print_line(log))

    @check_min_cli_version
    def get_deployment_info_by_fqn(self, deployment_fqn: str) -> DeploymentFqnResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/deployment"
        res = requests.get(
            url, headers=self._get_header(), params={"fqn": deployment_fqn}
        )
        res = request_handling(res)
        return DeploymentFqnResponse.parse_obj(res)

    @check_min_cli_version
    def get_application_info_by_fqn(
        self, application_fqn: str
    ) -> ApplicationFqnResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/fqn/app"
        res = requests.get(
            url, headers=self._get_header(), params={"fqn": application_fqn}
        )
        res = request_handling(res)
        return ApplicationFqnResponse.parse_obj(res)

    @check_min_cli_version
    def remove_application(self, application_id: str):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}"
        res = requests.delete(url, headers=self._get_header())
        response = request_handling(res)
        # TODO: Add pydantic here.
        return response

    @check_min_cli_version
    def get_application_info(self, application_id: str) -> Application:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}"
        res = requests.get(url, headers=self._get_header())
        response = request_handling(res)
        return Application.parse_obj(response)

    def list_job_runs(
        self,
        application_id: str,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        search_prefix: Optional[str] = None,
    ) -> List[JobRun]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/{application_id}/runs"
        params = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if search_prefix:
            params["searchPrefix"] = search_prefix
        res = requests.get(url, headers=self._get_header(), params=params)
        res = request_handling(res)
        return parse_obj_as(List[JobRun], res["data"])

    def get_job_run(
        self,
        application_id: str,
        job_run_name: str,
    ):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/{application_id}/runs/{job_run_name}"
        res = requests.get(url, headers=self._get_header())
        res = request_handling(res)
        return parse_obj_as(JobRun, res)

    def trigger_job(
        self,
        deployment_id: str,
        component_name: str,
        command: Optional[str] = None,
        params: Optional[Dict[str, str]] = None,
    ) -> TriggerJobResult:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/trigger"
        body = {
            "deploymentId": deployment_id,
            "componentName": component_name,
            "input": {},
        }
        if command:
            body["input"]["command"] = command
        if params:
            body["input"]["params"] = params
        res = requests.post(url, json=body, headers=self._get_header())
        response = request_handling(res)
        return TriggerJobResult.parse_obj(response)

    @check_min_cli_version
    def get_docker_registry_creds(
        self, docker_registry_fqn: str, cluster_id: str
    ) -> DockerRegistryCredentials:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/docker-registry/creds"
        res = requests.get(
            url,
            headers=self._get_header(),
            params={
                "fqn": docker_registry_fqn,
                "clusterId": cluster_id,
            },
        )
        response = request_handling(res)
        return DockerRegistryCredentials.parse_obj(response)

    @check_min_cli_version
    def create_repo_in_registry(
        self, docker_registry_fqn: str, workspace_fqn: str, application_name: str
    ) -> CreateDockerRepositoryResponse:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/docker-registry/create-repo"
        res = requests.post(
            url,
            headers=self._get_header(),
            data={
                "fqn": docker_registry_fqn,
                "workspaceFqn": workspace_fqn,
                "applicationName": application_name,
            },
        )
        response = request_handling(res)
        return CreateDockerRepositoryResponse.parse_obj(response)

    @check_min_cli_version
    def list_applications(
        self,
        application_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        application_name: Optional[str] = None,
    ) -> List[Application]:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/app"
        params = {}
        if application_id:
            params["applicationId"] = application_id
        if workspace_id:
            params["workspaceId"] = workspace_id
        if application_name:
            params["applicationName"] = application_name
        res = requests.get(url, params=params, headers=self._get_header())
        response = request_handling(res)
        return parse_obj_as(List[Application], response)

    @check_min_cli_version
    def list_versions(
        self,
        application_id: str,
        deployment_version: Optional[int] = None,
        deployment_id: Optional[str] = None,
    ) -> List[Deployment]:
        url = (
            f"{self._api_server_url}/{VERSION_PREFIX}/app/{application_id}/deployments"
        )
        params = {}
        if deployment_version:
            params["version"] = deployment_version
        if deployment_id:
            params["deploymentId"] = deployment_id
        res = requests.get(url, params=params, headers=self._get_header())
        response = request_handling(res)
        return parse_obj_as(List[Deployment], response)

    @check_min_cli_version
    def get_token_from_api_key(self, api_key: str) -> Token:
        url = f"{self._api_server_url}/{VERSION_PREFIX}/oauth/api-key/token"
        data = {"apiKey": api_key}
        res = requests.get(url, params=data)
        res = request_handling(res)
        return Token.parse_obj(res)

    def terminate_job_run(
        self,
        deployment_id: str,
        job_run_name: str,
        callback=OutputCallBack(),
    ):
        url = f"{self._api_server_url}/{VERSION_PREFIX}/jobs/terminate?deploymentId={deployment_id}&jobRunName={job_run_name}"
        body = {
            "deploymentId": deployment_id,
            "jobRunName": job_run_name,
        }
        res = requests.post(url, json=body, headers=self._get_header())
        res = request_handling(res)

        return res



# generated by datamodel-codegen:
#   filename:  application.json
#   timestamp: 2024-02-15T05:19:20+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Union

from servicefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    PositiveFloat,
    PositiveInt,
    confloat,
    conint,
    constr,
)


class AWSAccessKeyAuth(BaseModel):
    aws_access_key_id: str = Field(
        ..., description="+label=AWS Access Key ID\n+usage=AWS Access Key ID\n+sort=1"
    )
    aws_secret_access_key: str = Field(
        ...,
        description="+label=AWS Secret Access Key\n+usage=AWS Secret Access Key for the user to authenticate with\n+sort=2",
    )
    aws_session_token: Optional[str] = Field(
        None,
        description="+label=AWS Session Token\n+usage=AWS Session Token, only required when using temporary credentials\n+sort=3",
    )


class AWSInferentia(BaseModel):
    type: constr(regex=r"^aws_inferentia$") = Field(
        ..., description="+value=aws_inferentia"
    )
    name: Optional[str] = Field(
        None,
        description="+label=Inferentia accelerator name\n+usage=Name of the AWS Inferentia Accccelerator. One of [INF1, INF2].\nThis field is required for Node Selector and can be ignored in Nodepool Selector.",
    )
    count: conint(ge=1, le=16) = Field(
        ...,
        description="+label=Count\n+usage=Count of Inferentia accelerator chips to provide to the application",
    )


class ArtifactsCacheVolume(BaseModel):
    """
    +docs=Describes the volume that will be used to cache the models
    +label=Artifacts Cache Volume
    """

    storage_class: str = Field(
        ...,
        description="+label=Storage Class\n+usage=Storage class of the Volume where artifacts will be cached",
    )
    cache_size: conint(ge=1, le=1000) = Field(
        200,
        description="+label=Cache Size (GB)\n+usage=Size of the Volume (in GB) where artifacts will be cached. Should be greater than twice the size of artifacts getting cached",
    )


class AsyncProcessorSidecar(BaseModel):
    destination_url: str = Field(
        ...,
        description="+label=Destination URL\n+usage=URL for the processor to invoke",
    )
    request_timeout: conint(ge=1) = Field(
        10,
        description="+label=Request Timeout Seconds\n+usage=Timeout for the invoke request in seconds",
    )
    sidecar_image: Optional[str] = Field(
        None,
        description="+label=Sidecar Image\n+usage=Image for the processor sidecar (This field will be deprecated in the future)",
    )


class BaseAutoscaling(BaseModel):
    min_replicas: conint(ge=0) = Field(
        1,
        description="+label=Minimum replicas\n+usage=Minimum number of replicas to keep available\n+sort=1",
    )
    max_replicas: conint(ge=1, le=500) = Field(
        ...,
        description="+label=Maximum replicas\n+usage=Maximum number of replicas allowed for the component.\n+sort=2",
    )
    polling_interval: conint(ge=0) = Field(
        30,
        description="+label=Polling Interval\n+usage=This is the interval to check each trigger on.",
    )
    cooldown_period: conint(ge=0) = Field(
        300,
        description="+label=Cooldown Period\n+usage=The period to wait after the last trigger reported active before scaling the resource back to 0.",
    )


class BasicAuthCreds(BaseModel):
    """
    +label=Username and password for service auth
    """

    type: constr(regex=r"^basic_auth$") = Field(..., description="+value=basic_auth")
    username: constr(regex=r"^[a-z0-9-@\.]{1,64}$") = Field(
        ...,
        description="+label=Username for service auth\n+message=Upto 64 lower case alphanumeric character long\n+sort=1",
    )
    password: constr(regex=r"^[a-zA-Z0-9\.!@#$%^&*()_+=\-]{1,64}$") = Field(
        ...,
        description="+label=Password for service auth\n+message=Password should not be more than 64 characters\n+sort=2",
    )


class BlueGreen(BaseModel):
    """
    +docs=This strategy brings up the new release completely before switching the complete load to the new release.
    This minimizes the time that two versions are serving traffic at the same time.
    +label=Blue Green strategy
    """

    type: constr(regex=r"^blue_green$") = Field(..., description="+value=blue_green")
    enable_auto_promotion: bool = Field(
        False,
        description="+docs=Promote the new release to handle the complete traffic. A manual promotion would be needed if this is disabled\n+label=Auto-promotion",
    )
    auto_promotion_seconds: conint(ge=0) = Field(
        30,
        description="+docs=Promote the new release to handle the complete traffic after waiting for these many seconds\n+label=Auto-promotion seconds",
    )


class CPUUtilizationMetric(BaseModel):
    type: constr(regex=r"^cpu_utilization$") = Field(
        ..., description="+value=cpu_utilization"
    )
    value: conint(lt=100, gt=0) = Field(
        ...,
        description="+label=CPU utilization %\n+usage=Percentage of cpu request averaged over all replicas which the autoscaler should try to maintain",
    )


class CanaryStep(BaseModel):
    weight_percentage: conint(ge=0, le=100) = Field(
        ...,
        description="+docs=Percentage of total traffic to be shifted to the canary release.\nThe rest will continue to go to the existing deployment\n+label=Canary weight percentage\n+unit=%\n+placeholder=Weight",
    )
    pause_duration: conint(ge=0) = Field(
        30,
        description="+docs=Duration for which to pause the release. The release process will wait for these seconds before proceeding to the next step.\nIf this is not set, the step will pause indefinitely on this step\n+label=Pause duration\n+unit=seconds\n+placeholder=Duration",
    )


class CodeserverImage(BaseModel):
    """
    +usage=Codeserver with persistent environment (Python 3.11.6)
    """

    type: constr(regex=r"^codeserver$") = Field(..., description="+value=codeserver")
    enable_sudo: bool = Field(
        True,
        description="+label=Enable root access to the container\n+usage=Changes made to the root directory `/` will not be persisted across notebook restarts",
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class CronMetric(BaseModel):
    type: constr(regex=r"^cron$") = Field(..., description="+value=cron")
    desired_replicas: Optional[conint(ge=1)] = Field(
        None,
        description="+label=Desired Replicas\n+usage=Desired number of replicas during the given interval. Default value is max_replicas.",
    )
    start: str = Field(
        ...,
        description="+label=Start Schedule\n+docs=Cron expression indicating the start of the cron schedule.\n+usage=Cron expression indicating the start of the cron schedule.\n```\n* * * * *\n| | | | |\n| | | | |___ day of week (0-6) (Sunday is 0)\n| | | |_____ month (1-12)\n| | |_______ day of month (1-31)\n| |_________ hour (0-23)\n|___________ minute (0-59)\n```",
    )
    end: str = Field(
        ...,
        description="+label=End Schedule\n+docs=Cron expression indicating the end of the cron schedule.\n+usage=Cron expression indicating the end of the cron schedule.\n```\n* * * * *\n| | | | |\n| | | | |___ day of week (0-6) (Sunday is 0)\n| | | |_____ month (1-12)\n| | |_______ day of month (1-31)\n| |_________ hour (0-23)\n|___________ minute (0-59)\n```",
    )
    timezone: str = Field(
        "UTC",
        description='+usage=Timezone against which the cron schedule will be calculated, e.g. "Asia/Tokyo". Default is machine\'s local time.\nhttps://docs.truefoundry.com/docs/list-of-supported-timezones',
    )


class CustomCodeserverImage(BaseModel):
    """
    +usage=User supplied docker image URI for vscode server
    """

    type: constr(regex=r"^customcodeserver$") = Field(
        ..., description="+value=customcodeserver"
    )
    image_uri: str = Field(
        ...,
        description="+label=Image URI\n+usage=The image URI. Specify the name of the image and the tag.\nIf the image is in Dockerhub, you can skip registry-url (for e.g. `tensorflow/tensorflow`).\nYou can use an image from a private registry using Advanced fields\n+placeholder=registry-url/account/image:version",
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class CustomNotebookImage(BaseModel):
    """
    +usage=User supplied docker image URI for jupyter notebook
    """

    type: constr(regex=r"^customnotebook$") = Field(
        ..., description="+value=customnotebook"
    )
    image_uri: str = Field(
        ...,
        description="+label=Image URI\n+usage=The image URI. Specify the name of the image and the tag.\nIf the image is in Dockerhub, you can skip registry-url (for e.g. `tensorflow/tensorflow`).\nYou can use an image from a private registry using Advanced fields\n+placeholder=registry-url/account/image:version",
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class CustomSSHServerImage(BaseModel):
    """
    +usage=User supplied docker image URI for ssh server
    """

    type: constr(regex=r"^custom-ssh-server$") = Field(
        ..., description="+value=custom-ssh-server"
    )
    image_uri: str = Field(
        ...,
        description="+label=Image URI\n+usage=The image URI. Specify the name of the image and the tag.\nIf the image is in Dockerhub, you can skip registry-url (for e.g. `tensorflow/tensorflow`).\nYou can use an image from a private registry using Advanced fields\n+placeholder=registry-url/account/image:version",
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class DockerFileBuild(BaseModel):
    """
    +docs=Describes that we are using a dockerfile to build our image
    +label=Docker File (I already have Docker File)
    +icon=fa-brands fa-docker:#0db7ed
    """

    type: constr(regex=r"^dockerfile$") = Field(..., description="+value=dockerfile")
    dockerfile_path: str = Field(
        "./Dockerfile",
        description="+label=Path to Dockerfile\n+usage=The file path of the Dockerfile relative to project root path.",
    )
    build_context_path: str = Field(
        "./",
        description="+label=Path to build context\n+usage=Build context path for the Dockerfile relative to project root path.",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Command Override\n+usage=Override the command to run when the container starts\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )
    build_args: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Build Args\n+usage=Build arguments to pass to `docker build`.",
    )


class Endpoint(BaseModel):
    host: constr(
        regex=r"^((([a-zA-Z0-9\-]{1,63}\.)([a-zA-Z0-9\-]{1,63}\.)*([A-Za-z]{1,63}))|(((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))$"
    ) = Field(
        ...,
        description="+usage=Host e.g. ai.example.com, app.truefoundry.com\n+message=Upto 253 characters, each part of host should be at most 63 characters long, can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts must be separated by periods (.)",
    )
    path: Optional[
        constr(regex=r"^(/([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_\.]*[a-zA-Z0-9]))*/$")
    ] = Field(
        None,
        description="+usage=Path e.g. /v1/api/ml/, /v2/docs/\n+message=Should begin and end with a forward slash (/). Each part can can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts should be separated by forward slashes (/)",
    )


class GitSource(BaseModel):
    """
    +docs=Describes that we are using code stored in a git repository to build our image
    +label=Git Source
    +icon=fa-solid fa-code-branch:black
    +sort=300
    """

    type: constr(regex=r"^git$") = Field(..., description="+value=git")
    repo_url: constr(
        regex=r"^(http(s?)://)(github\.com|(.+@)*bitbucket\.org|gitlab\.com).*$"
    ) = Field(
        ...,
        description="+label=Repo URL\n+usage=The repository URL.\n+sort=1\n+message=Needs to be a valid Github, Bitbucket or Gitlab link",
    )
    ref: str = Field(
        ..., description="+label=Commit SHA\n+usage=The commit SHA.\n+sort=2"
    )
    branch_name: Optional[str] = Field(
        None,
        description="+label=Branch Name\n+usage=Selecting branch will select latest commit SHA of the branch.\n+sort=3",
    )


class HelmRepo(BaseModel):
    type: constr(regex=r"^helm-repo$") = Field(..., description="+value=helm-repo")
    repo_url: constr(
        regex=r"^(((https?|wss):\/\/)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&\/=]*))$"
    ) = Field(
        ...,
        description="+label=Helm repository URL\n+sort=1\n+message=Needs to be a valid URL.",
    )
    chart: str = Field(
        ..., description="+label=Chart name\n+sort=2\n+usage=The helm chart name"
    )
    version: str = Field(
        ..., description="+label=Version\n+sort=3\n+usage=Helm chart version"
    )


class HttpProbe(BaseModel):
    """
    +docs=Describes the Instructions for assessing container health by executing an HTTP GET request.
    To learn more you can go [here](https://docs.truefoundry.com/docs/add-health-checks-to-deployments)
    +label=Instructions for assessing container health by executing an HTTP GET request.
    """

    type: constr(regex=r"^http$") = Field(..., description="+sort=1\n+value=http")
    path: str = Field(
        ...,
        description="+usage=The endpoint, relative to the port, to which the HTTP GET request should be directed.\n+sort=2",
    )
    port: conint(ge=0, le=65535) = Field(
        ...,
        description="+usage=The TCP socket within the container to which the HTTP GET request should be directed.\n+sort=3",
    )
    host: Optional[str] = Field(
        None,
        description="+sort=4\n+usage=Host name to connect to, defaults to the pod IP",
    )
    scheme: str = Field(
        "HTTP", description="+sort=5\n+usage=Scheme to use for connecting to the host"
    )


class HuggingfaceArtifactSource(BaseModel):
    """
    +docs=Input for Artifact from Huggingface Model Hub
    +label=Huggingface Model Source
    """

    type: constr(regex=r"^huggingface-hub$") = Field(
        ..., description="+value=huggingface-hub"
    )
    model_id: str = Field(
        ...,
        description="+label=Model ID\n+usage=Model ID of the artifact to be downloaded",
    )
    revision: str = Field(
        ...,
        description="+label=Revision\n+usage=Revision of the artifact to be downloaded",
    )
    ignore_patterns: List[str] = Field(
        ["*.h5", "*.ot", "*.tflite", "*.msgpack"],
        description="+label=Ignore Patterns\n+usage=List of patterns to ignore while downloading the artifact",
    )
    download_path_env_variable: str = Field(
        ...,
        description="+label=Download Path Environment Variable\n+usage=Environment variable which will contain the download path of the artifact",
    )


class Image(BaseModel):
    """
    +docs=Describes that we are using a pre-built image stored in a Docker Image registry
    +label=Docker Image (Deploy an existing image)
    +icon=fa-brands fa-docker:#0db7ed
    """

    type: constr(regex=r"^image$") = Field(..., description="+value=image")
    image_uri: constr(regex=r"^\S*$") = Field(
        ...,
        description="+label=Image URI\n+usage=The image URI. Specify the name of the image and the tag.\nIf the image is in Dockerhub, you can skip registry-url (for e.g. `tensorflow/tensorflow`).\nYou can use an image from a private registry using Advanced fields\n+placeholder=registry-url/account/image:version (e.g. docker.io/tensorflow/tensorflow)",
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Command Override\n+usage=Override the command to run when container starts.\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )


class KafkaMetricConfig(BaseModel):
    type: constr(regex=r"^kafka$") = Field(..., description="+value=kafka")
    lag_threshold: PositiveInt = Field(
        ...,
        description="+label=Lag Threshold\n+usage=Upper limit of the number of backlog messages the auto-scaler will try to maintain per replica. If you set this number to 10 and have 30 messages in the stream and one replica, the auto-scaler will scale the number of replicas to 3.",
    )


class KafkaSASLAuth(BaseModel):
    username: str = Field(
        ...,
        description="+label=Username\n+usage=Username for SASL authentication\n+sort=1",
    )
    password: str = Field(
        ...,
        description="+label=Password\n+usage=Password for SASL authentication\n+sort=2",
    )


class Kustomize(BaseModel):
    patch: Optional[Dict[str, Any]] = Field(
        None,
        description="+label=Patch\n+usage=Content of kustomization.yaml to perform kustomize operation. Please do not include the `resources` section. It is filled in automatically",
    )
    additions: Optional[List[Dict[str, Any]]] = Field(
        None,
        description="+label=Additional Manifests\n+usage=Additional kubernetes manifests to be included in the application",
    )


class LocalSource(BaseModel):
    """
    +docs=Describes that we are using code stored in a local developement environment to build our image
    +label=Local
    +icon=fa-folder:black
    +sort=100
    """

    type: constr(regex=r"^local$") = Field(..., description="+value=local")
    project_root_path: str = Field("./", description="+usage=Local project root path.")
    local_build: bool = Field(True, description="run docker build locally")


class Manual(BaseModel):
    """
    +docs=Describes that we are going to manually trigger our job.
    +label=Manual
    +usage=Trigger the job manually. [Docs](https://docs.truefoundry.com/docs/more-configurations)
    """

    type: constr(regex=r"^manual$") = Field(..., description="+value=manual")


class NATSMetricConfig(BaseModel):
    type: constr(regex=r"^nats$") = Field(..., description="+value=nats")
    lag_threshold: PositiveInt = Field(
        ...,
        description="+label=Lag Threshold\n+usage=Upper limit of the number of backlog messages the auto-scaler will try to maintain per replica. If you set this number to 10 and have 30 messages in the stream and one replica, the auto-scaler will scale the number of replicas to 3.",
    )


class NATSUserPasswordAuth(BaseModel):
    """
    +docs=NATS User Password Authentication
    +label=NATS User Password Authentication
    """

    account_name: str = Field(
        "$G",
        description="+label=Account Name\n+usage=Name of the NATS account\n+sort=1",
    )
    user: str = Field(
        ..., description="+label=User\n+usage=User for NATS authentication\n+sort=2"
    )
    password: str = Field(
        ...,
        description="+label=Password\n+usage=Password for NATS authentication\n+sort=3",
    )


class CapacityType(str, Enum):
    """
    +label=Capacity Type
    +usage=Configure what type of nodes to run the app. By default no placement logic is applied.
    "spot_fallback_on_demand" will try to place the application on spot nodes but will fallback to on-demand when spot nodes are not available.
    "spot" will strictly place the application on spot nodes.
    "on_demand" will strictly place the application on on-demand nodes.
    """

    spot_fallback_on_demand = "spot_fallback_on_demand"
    spot = "spot"
    on_demand = "on_demand"


class NodeSelector(BaseModel):
    """
    +label=Node selector
    +usage=Constraints to select a Node - Specific GPU / Instance Families, On-Demand/Spot.
    """

    type: constr(regex=r"^node_selector$") = Field(
        ..., description="+value=node_selector"
    )
    gpu_type: Optional[str] = Field(
        None,
        description="+label=GPU Type\n+usage=Name of the Nvidia GPU. One of [P4, P100, V100, T4, A10G, A100_40GB, A100_80GB]\nOne instance of the card contains the following amount of memory -\nP4: 8 GB, P100: 16 GB, V100: 16 GB, T4: 16 GB, A10G: 24 GB, A100_40GB: 40GB, A100_80GB: 80 GB",
    )
    instance_families: Optional[List[str]] = Field(
        None,
        description="+label=Instance family\n+usage=Instance family of the underlying machine to use. Multiple instance families can be supplied.\nThe workload is guaranteed to be scheduled on one of them.",
    )
    capacity_type: Optional[CapacityType] = Field(
        None,
        description='+label=Capacity Type\n+usage=Configure what type of nodes to run the app. By default no placement logic is applied.\n"spot_fallback_on_demand" will try to place the application on spot nodes but will fallback to on-demand when spot nodes are not available.\n"spot" will strictly place the application on spot nodes.\n"on_demand" will strictly place the application on on-demand nodes.',
    )


class NodepoolSelector(BaseModel):
    """
    +label=Nodepool selector
    +usage=Specify one or more nodepools to run your application on.
    """

    type: constr(regex=r"^nodepool_selector$") = Field(
        ..., description="+value=nodepool_selector"
    )
    nodepools: Optional[List[str]] = Field(
        None,
        description="+label=Nodepools\n+usage=Nodepools where you want to run your workload. Multiple nodepools can be selected.\n The workload is guaranteed to be scheduled on one of the nodepool",
    )


class NvidiaGPU(BaseModel):
    type: constr(regex=r"^nvidia_gpu$") = Field(..., description="+value=nvidia_gpu")
    name: Optional[str] = Field(
        None,
        description="+label=GPU Name\n+usage=Name of the Nvidia GPU. One of [P4, P100, V100, T4, A10G, A100_40GB, A100_80GB]\nThis field is required for Node Selector and can be ignored in Nodepool Selector.\nOne instance of the card contains the following amount of memory -\nP4: 8 GB, P100: 16 GB, V100: 16 GB, T4: 16 GB, A10G: 24 GB, A100_40GB: 40GB, A100_80GB: 80 GB",
    )
    count: conint(ge=1, le=16) = Field(
        ...,
        description="+label=GPU Count\n+usage=Count of GPUs to provide to the application\nNote the exact count and max count available for a given GPU type depends on cloud provider and cluster type.",
    )


class OCIRepo(BaseModel):
    """
    +label=OCIRepo
    """

    type: constr(regex=r"^oci-repo$") = Field(..., description="+value=oci-repo")
    oci_chart_url: constr(
        regex=r"^(((oci):\/\/)?(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}(?:[-a-zA-Z0-9()@:%_\+.~#?&\/=]*))$"
    ) = Field(..., description="+label=OCI chart URL\n+message=Need to be a valid URL.")
    version: str = Field(..., description="+label=Version\n+usage=Helm chart version")


class ParamType(str, Enum):
    string = "string"
    ml_repo = "ml_repo"


class Param(BaseModel):
    name: constr(regex=r"^[a-z][a-z0-9\-_]{0,30}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the param\n+message=name can contain lower case alphabets, digits, underscore (_) and hypen (-). It can be 32 characters long, should start with an alphabet, and should end with either an alphabet or digit.",
    )
    description: Optional[constr(regex=r"^.{1,127}$")] = Field(
        None,
        description="+usage=Description of param\n+message=description cannot be longer than 127 characters",
    )
    default: Optional[constr(regex=r"^.{0,127}$")] = Field(
        None,
        description="+usage=Default value or placeholder\n+message=default value cannot be longer than 127 characters\n+label=Default value",
    )
    param_type: ParamType = "string"


class Protocol(str, Enum):
    """
    +usage=Protocol for the port.
    """

    TCP = "TCP"
    UDP = "UDP"


class AppProtocol(str, Enum):
    """
    +label=Application Protocol
    +usage=Application Protocol for the port.
    Select the application protocol used by your service. For most use cases, this should be `http`(HTTP/1.1).
    If you are running a gRPC server, select the `grpc` option.
    This is only applicable if `expose=true`.
    """

    http = "http"
    grpc = "grpc"
    tcp = "tcp"


class Port(BaseModel):
    """
    +docs=Describes the ports the service should be exposed to.
    """

    port: conint(ge=1, le=65535) = Field(
        80, description="+usage=Port number to expose."
    )
    protocol: Protocol = Field("TCP", description="+usage=Protocol for the port.")
    expose: bool = Field(True, description="+usage=Expose the port")
    app_protocol: Optional[AppProtocol] = Field(
        None,
        description="+label=Application Protocol\n+usage=Application Protocol for the port.\nSelect the application protocol used by your service. For most use cases, this should be `http`(HTTP/1.1).\nIf you are running a gRPC server, select the `grpc` option.\nThis is only applicable if `expose=true`.",
    )
    host: Optional[
        constr(
            regex=r"^((([a-zA-Z0-9\-]{1,63}\.)([a-zA-Z0-9\-]{1,63}\.)*([A-Za-z]{1,63}))|(((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))$"
        )
    ] = Field(
        None,
        description="+usage=Host e.g. ai.example.com, app.truefoundry.com\n+message=Upto 253 characters, each part of host should be at most 63 characters long, can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts must be separated by periods (.)",
    )
    path: Optional[
        constr(regex=r"^(/([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_\.]*[a-zA-Z0-9]))*/$")
    ] = Field(
        None,
        description="+usage=Path e.g. /v1/api/ml/, /v2/docs/\n+message=Should begin and end with a forward slash (/). Each part can can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts should be separated by forward slashes (/)",
    )
    rewrite_path_to: Optional[
        constr(regex=r"^(/([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_\.]*[a-zA-Z0-9]))*/$")
    ] = Field(
        None,
        description="+label=Rewrite Path to\n+usage=Rewrite the path prefix to a different path.\nIf `path` is `/v1/api` and `rewrite_path_to` is `/api`. The URI in the HTTP request `http://0.0.0.0:8080/v1/api/houses` will be rewritten to `http://0.0.0.0:8080/api/houses` before the request is forwarded your service.\nDefaults to `/`.\nThis is only applicable if `path` is given.\n+message=Should begin and end with a forward slash (/). Each part can can contain alphabets, digits and hypen, must begin and end with an alphanumeric characters. Parts should be separated by forward slashes (/)",
    )
    auth: Optional[BasicAuthCreds] = None


class PythonBuild(BaseModel):
    """
    +docs=Describes that we are using python to build a container image with a specific python version and pip packages installed.
    +label=Python Code (I don't have Dockerfile)
    +icon=fa-brands fa-python:#306998
    """

    type: constr(regex=r"^tfy-python-buildpack$") = Field(
        ..., description="+value=tfy-python-buildpack"
    )
    python_version: constr(regex=r"^\d+(\.\d+){1,2}([\-\.a-z0-9]+)?$") = Field(
        "3.9",
        description="+label=Python version\n+usage=Python version to run your application. Should be one of the tags listed on [Official Python Docker Page](https://hub.docker.com/_/python)\n+message=Please enter a valid Python version tag",
    )
    build_context_path: str = Field(
        "./",
        description="+label=Path to build context\n+usage=Build path relative to project root path.",
    )
    requirements_path: Optional[str] = Field(
        None,
        description="`Path to build context`\n+label=Path to requirements\n+usage=Path to `requirements.txt` relative to\n`Path to build context`",
    )
    pip_packages: Optional[List[str]] = Field(
        None,
        description='+label=Pip packages to install\n+usage=Define pip package requirements.\nIn Python/YAML E.g. ["fastapi>=0.90,<1.0", "uvicorn"]\n+placeholder=Enter a pip package name E.g. fastapi>=0.90,<1.0',
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    command: Union[str, List[str]] = Field(
        ...,
        description="Command will be set as the Entrypoint of the generated\nimage.\n+label=Command\n+usage=Command to run when the container starts.\nCommand will be set as the Entrypoint of the generated image.\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )
    cuda_version: Optional[
        constr(
            regex=r"^((\d+\.\d+(\.\d+)?-cudnn\d+-(runtime|devel)-ubuntu\d+\.\d+)|11\.0-cudnn8|11\.1-cudnn8|11\.2-cudnn8|11\.3-cudnn8|11\.4-cudnn8|11\.5-cudnn8|11\.6-cudnn8|11\.7-cudnn8|11\.8-cudnn8|12\.0-cudnn8|12\.1-cudnn8|12\.2-cudnn8)$"
        )
    ] = Field(
        None,
        description="+label=CUDA Version\n+usage=Version of CUDA Toolkit and CUDNN to install in the image\nThese combinations are based off of publically available docker images on docker hub\nYou can also specify a valid tag of the form {cuda_version_number}-cudnn{cudnn_version_number}-{runtime|devel}-ubuntu{ubuntu_version}\nRefer https://hub.docker.com/r/nvidia/cuda/tags for valid set of values\nNote: We use deadsnakes ubuntu ppa to add Python that currently supports only Ubuntu 18.04, 20.04 and 22.04",
    )


class RPSMetric(BaseModel):
    type: constr(regex=r"^rps$") = Field(..., description="+value=rps")
    value: PositiveFloat = Field(
        ...,
        description="+label=Requests per second\n+usage=Average request per second averaged over all replicas that autoscaler should try to maintain",
    )


class RemoteSource(BaseModel):
    """
    +docs=Describes that we are using code stored in a remote respository to build our image
    +label=S3
    +icon=fa-brands fa-aws:black
    +sort=200
    """

    type: constr(regex=r"^remote$") = Field(..., description="+value=remote")
    remote_uri: str = Field(
        ..., description="+docs=Remote repository URI\n+label=Remote URI"
    )


class Resources(BaseModel):
    """
    +docs=Describes the resource constraints for the application so that it can be deployed accordingly on the cluster
    To learn more you can go [here](https://docs.truefoundry.com/docs/resources)
    +icon=fa-microchip
    +label=Resources
    +usage=Set resource constraints for the application. [Docs](https://docs.truefoundry.com/docs/resources)
    """

    cpu_request: confloat(ge=0.001, le=256.0) = Field(
        0.2,
        description="+label=CPU Request\n+sort=1\n+usage=Requested CPU which determines the minimum cost incurred. The CPU usage can exceed the requested\namount, but not the value specified in the limit. 1 CPU means 1 CPU core. Fractional CPU can be requested\nlike `0.5` or `0.05`",
    )
    cpu_limit: confloat(ge=0.001, le=256.0) = Field(
        0.5,
        description="+label=CPU Limit\n+usage=CPU limit beyond which the usage cannot be exceeded. 1 CPU means 1 CPU core. Fractional CPU can be requested\nlike `0.5`. CPU limit should be >= cpu request.\n+sort=2",
    )
    memory_request: conint(ge=1, le=2000000) = Field(
        200,
        description="+label=Memory Request\n+usage=Requested memory which determines the minimum cost incurred. The unit of memory is in megabytes(MB).\nSo 1 means 1 MB and 2000 means 2GB.\n+sort=3",
    )
    memory_limit: conint(ge=1, le=2000000) = Field(
        500,
        description="+label=Memory Limit\n+usage=Memory limit after which the application will be killed with an OOM error. The unit of memory is\nin megabytes(MB). So 1 means 1 MB and 2000 means 2GB. MemoryLimit should be greater than memory request.\n+sort=4",
    )
    ephemeral_storage_request: conint(ge=1, le=2000000) = Field(
        1000,
        description="+label=Storage Request\n+usage=Requested disk storage. The unit of memory is in megabytes(MB).\nThis is ephemeral storage and will be wiped out on pod restarts or eviction\n+sort=5",
    )
    ephemeral_storage_limit: conint(ge=1, le=2000000) = Field(
        2000,
        description="+label=Storage Limit\n+usage=Disk storage limit. The unit of memory is in megabytes(MB). Exceeding this limit will result in eviction.\nIt should be greater than the request. This is ephemeral storage and will be wiped out on pod restarts or eviction\n+sort=6",
    )
    gpu_count: conint(ge=0, le=16) = Field(
        0,
        description="+label=GPU Count\n+usage=Count of GPUs to provide to the application\nNote the exact count and max count available for a given GPU type depends on cloud provider and cluster type.",
    )
    shared_memory_size: Optional[conint(ge=64, le=32000)] = Field(
        None,
        description="+label=Shared Memory Size (MB)\n+usage=Define the shared memory requirements for your workload. Machine learning libraries like Pytorch can use Shared Memory\nfor inter-process communication. If you use this, we will mount a `tmpfs` backed volume at the `/dev/shm` directory.\nAny usage will also count against the workload's memory limit (`resources.memory_limit`) along with your workload's memory usage.\nIf the overall usage goes above `resources.memory_limit` the user process may get killed.\nShared Memory Size cannot be more than the defined Memory Limit for the workload.",
    )
    node: Optional[Union[NodeSelector, NodepoolSelector]] = Field(
        None,
        description="+label=Node\n+usage=This field determines how the underlying node resource is to be utilized",
    )
    devices: Optional[List[Union[NvidiaGPU, AWSInferentia]]] = Field(
        None,
        description="+label=Devices\n+usage=Define custom device or accelerator requirements for your workload. We currently support NVIDIA GPUs and AWS Inferentia Accelerators.",
    )


class Rolling(BaseModel):
    """
    +docs=This strategy updates the pods in a rolling fashion such that a subset of the
    total pods are replaced with new version at one time.
    A commonly used strategy can be to have maxUnavailablePercentage close to 0 so that there
    is no downtime and keep the maxSurgePercentage to around 25%. If you are anyways running
    a large number of pods, the service can often tolerate a few pods going down - so you
    max maxUnavailablePercentage = 10 and maxSurgePercentage=0. You can read about it more
    [here](https://spot.io/resources/kubernetes-autoscaling/5-kubernetes-deployment-strategies-roll-out-like-the-pros/)
    +label=Rolling update strategy
    """

    type: constr(regex=r"^rolling_update$") = Field(
        ..., description="+value=rolling_update"
    )
    max_unavailable_percentage: conint(ge=0, le=100) = Field(
        25,
        description="+label=Max unavailable(%)\n+usage=Percentage of total replicas that can be brought down at one time.\nFor a value of 25 when replicas are set to 12 this would mean minimum (25% of 12) = 3 pods might be unavailable during the deployment.\nSetting this to a higher value can help in speeding up the deployment process.",
    )
    max_surge_percentage: conint(ge=0, le=100) = Field(
        0,
        description="+label=Max Surge(%)\n+usage=Percentage of total replicas of updated image that can be brought up over the total replicas count.\nFor a value of 25 when replicas are set to 12 this would mean (12+(25% of 12) = 15) pods might be running at one time.\nSetting this to a higher value can help in speeding up the deployment process.",
    )


class SQSInputConfig(BaseModel):
    """
    +docs=Describes the configuration for the input SQS worker
    +label=SQS Input Config
    """

    type: constr(regex=r"^sqs$") = Field(..., description="+value=sqs")
    queue_url: str = Field(
        ...,
        description="+label=Queue URL\n+usage=AWS SQS Queue URL of Subscriber\n+sort=1",
    )
    region_name: str = Field(
        ..., description="+label=Region Name\n+usage=AWS Region Name\n+sort=2"
    )
    visibility_timeout: conint(le=43200, gt=0) = Field(
        ...,
        description="+label=Visibility Timeout (seconds)\n+usage=A period during which Amazon SQS prevents all consumers from receiving and processing the message. If one message takes 5 seconds to process, you can set this number to 7 or any number higher than 5. This will ensure that while the message is being processed, it will not be available to other replicas. For more information, see [here](https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-visibility-timeout.html)\n+sort=3",
    )
    wait_time_seconds: conint(ge=1, le=20) = Field(
        19,
        description="+label=Wait Time Seconds\n+usage=Wait timeout for long polling. For more information, see [here](https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-short-and-long-polling.html)",
    )
    auth: AWSAccessKeyAuth


class SQSOutputConfig(BaseModel):
    """
    +docs=Describes the configuration for the output SQS worker
    +label=SQS Output Config
    """

    type: constr(regex=r"^sqs$") = Field(..., description="+value=sqs")
    queue_url: str = Field(
        ...,
        description="+label=Queue URL\n+usage=AWS SQS Queue URL of Publisher\n+sort=1",
    )
    region_name: str = Field(
        ..., description="+label=Region Name\n+usage=AWS Region Name\n+sort=2"
    )
    auth: AWSAccessKeyAuth


class SQSQueueMetricConfig(BaseModel):
    type: constr(regex=r"^sqs$") = Field(..., description="+value=sqs")
    queue_length: PositiveInt = Field(
        ...,
        description="+label=Queue length\n+usage=Upper limit of the number of backlog messages the auto-scaler will try to maintain per replica. If you set this number to 10 and have 30 messages in the queue and one replica, the auto-scaler will scale the number of replicas to 3.",
    )


class SSHServerImage(BaseModel):
    """
    +usage=Ssh Server with persistent environment (Python 3.11.6)
    """

    type: constr(regex=r"^ssh-server$") = Field(..., description="+value=ssh-server")
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class ConcurrencyPolicy(str, Enum):
    """
    +usage=Choose whether to allow this job to run while another instance of the job is running, or to replace the currently running instance. Allow
    will enable multiple instances of this job to run. Forbid will keep the current instance of the job running and stop a new instance from being run.
    Replace will terminate any currently running instance of the job and start a new one.
    """

    Forbid = "Forbid"
    Allow = "Allow"
    Replace = "Replace"


class Schedule(BaseModel):
    """
    +docs=Describes that we are going to schedule our job to run at a schedule, making our job a cron job.
    +label=Schedule
    +usage=Run the job on a schedule. [Docs](https://docs.truefoundry.com/docs/deploy-a-cron-job)
    """

    type: constr(regex=r"^scheduled$") = Field(..., description="+value=scheduled")
    schedule: str = Field(
        ...,
        description="+docs=Specify the schedule for this job to be run periodically in cron format. [Learn more](https://docs.truefoundry.com/docs/deploy-a-cron-job)\n+usage=Specify the schedule for this job to be run periodically in cron format.\n```\n* * * * *\n| | | | |\n| | | | |___ day of week (0-6) (Sunday is 0)\n| | | |_____ month (1-12)\n| | |_______ day of month (1-31)\n| |_________ hour (0-23)\n|___________ minute (0-59)\n```",
    )
    concurrency_policy: ConcurrencyPolicy = Field(
        "Forbid",
        description="+usage=Choose whether to allow this job to run while another instance of the job is running, or to replace the currently running instance. Allow\nwill enable multiple instances of this job to run. Forbid will keep the current instance of the job running and stop a new instance from being run.\nReplace will terminate any currently running instance of the job and start a new one.",
    )
    timezone: Optional[str] = Field(
        None,
        description='+usage=Timezone against which the cron schedule will be calculated, e.g. "Asia/Tokyo". Default is machine\'s local time.\nhttps://docs.truefoundry.com/docs/list-of-supported-timezones',
    )


class SecretMount(BaseModel):
    type: constr(regex=r"^secret$") = Field(..., description="+value=secret")
    mount_path: constr(regex=r"^\/(?:[^/\n]+\/*)*[^/\n]+(\.[^/\n]+)?$") = Field(
        ...,
        description="+label=File path\n+usage=Absolute file path where the file will be created.\n+message=Please enter a valid file path",
    )
    secret_fqn: constr(regex=r"^tfy-secret:\/\/.+:.+:.+$") = Field(
        ...,
        description="+label=Secret\n+usage=The Truefoundry secret whose value will be the file content.",
    )


class ServiceAutoscaling(BaseAutoscaling):
    metrics: Union[CPUUtilizationMetric, RPSMetric, CronMetric] = Field(
        ...,
        description="+label=Autoscaling metrics\n+usage=Metrics to use for the autoscaler\n+sort=4",
    )


class StringDataMount(BaseModel):
    type: constr(regex=r"^string$") = Field(..., description="+value=string")
    mount_path: constr(regex=r"^\/(?:[^/\n]+\/*)*[^/\n]+(\.[^/\n]+)?$") = Field(
        ...,
        description="+label=File Path\n+usage=Absolute file path where the file will be created.\n+message=Please enter a valid file path",
    )
    data: str = Field(..., description="+label=Data\n+usage=The file content.")


class TruefoundryArtifactSource(BaseModel):
    """
    +docs=Input for Artifact from Truefoundry Artifact Registry
    +label=Truefoundry Artifact Source
    """

    type: constr(regex=r"^truefoundry-artifact$") = Field(
        ..., description="+value=truefoundry-artifact"
    )
    artifact_version_fqn: str = Field(
        ...,
        description="+label=Artifact or Model Version FQN\n+usage=Artifact or Model Version FQN of the artifact to be downloaded",
    )
    download_path_env_variable: str = Field(
        ...,
        description="+label=Download Path Environment Variable\n+usage=Environment variable which will contain the download path of the artifact",
    )


class TruefoundryImageBase(BaseModel):
    """
    +usage=JupyterLab with persistent python environment (Python 3.11.6)
    """

    type: constr(regex=r"^truefoundrybase$") = Field(
        ..., description="+value=truefoundrybase"
    )
    enable_sudo: bool = Field(
        True,
        description="+label=Enable root access to the container\n+usage=Changes made to the root directory `/` will not be persisted across notebook restarts",
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class TruefoundryImageCuda1180(BaseModel):
    """
    +usage=JupyterLab with persistent python environment (Python 3.11.6, Cuda 11.8.0)
    """

    type: constr(regex=r"^truefoundrycuda1180$") = Field(
        ..., description="+value=truefoundrycuda1180"
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class TruefoundryImageCuda1211(BaseModel):
    """
    +usage=JupyterLab with persistent python environment (Python 3.11.6, Cuda 12.1.1)
    """

    type: constr(regex=r"^truefoundrycuda1211$") = Field(
        ..., description="+value=truefoundrycuda1211"
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class TruefoundryImageFull(BaseModel):
    """
    +usage=JupyterLab + Tensorflow 2.15.0 and Pytorch 2.1.1 with persistent environment (Python 3.11.6)
    """

    type: constr(regex=r"^truefoundryfull$") = Field(
        ..., description="+value=truefoundryfull"
    )
    enable_sudo: bool = Field(
        True,
        description="+label=Enable root access to the container\n+usage=Changes made to the root directory `/` will not be persisted across notebook restarts",
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=List of Debian packages to install.\n+usage=Debian packages to install via `apt get`.\nIn Python/YAML E.g. ["git", "ffmpeg", "htop"]\n+placeholder=Enter a debian package name E.g. ffmpeg',
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can use the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )


class VolumeBrowser(BaseModel):
    """
    +label=Volume Browser
    """

    username: constr(regex=r"^[a-z][a-z0-9]{1,8}[a-z0-9]$") = Field(
        ...,
        description="+message=3 to 10 lower case characters long alphanumeric word, may contain - in between, cannot start with a number.\n+usage=Username for logging in the volume browser.\n+sort=1",
    )
    password_secret_fqn: constr(regex=r"^tfy-secret:\/\/.+:.+:.+$") = Field(
        ...,
        description="+label=Password Secret FQN\n+usage=TFY Secret containing the password for logging in the volume browser.\n+sort=2",
    )
    endpoint: Endpoint


class VolumeMount(BaseModel):
    type: constr(regex=r"^volume$") = Field(..., description="+value=volume")
    mount_path: constr(regex=r"^\/(?:[^/\n]+\/*)*[^/\n]+(\.[^/\n]+)?$") = Field(
        ...,
        description="+label=Volume mount path\n+usage=Absolute file path where the volume will be mounted.\n+message=Please enter a valid mount path",
    )
    volume_fqn: constr(regex=r"^tfy-volume:\/\/.+:.+:.+$") = Field(
        ...,
        description="+label=Volume\n+usage=The Truefoundry volume that needs to be mounted.",
    )


class ArtifactsDownload(BaseModel):
    """
    +docs=Describes the configuration for the artifacts cache
    +label=Artifacts Download
    """

    cache_volume: Optional[ArtifactsCacheVolume] = None
    artifacts: List[
        Union[TruefoundryArtifactSource, HuggingfaceArtifactSource]
    ] = Field(
        ..., description="+label=Artifacts\n+usage=List of artifacts to be cached"
    )


class AsyncServiceAutoscaling(BaseAutoscaling):
    metrics: Union[
        SQSQueueMetricConfig, NATSMetricConfig, KafkaMetricConfig, CronMetric
    ] = Field(
        ...,
        description="+label=Autoscaling metrics\n+usage=Metrics to use for the autoscaler\n+sort=4",
    )


class BaseWorkbenchInput(BaseModel):
    """
    +docs=Describes the configuration for the service
    """

    name: constr(regex=r"^[a-z][a-z0-9\-]{1,30}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the workbench. This uniquely identifies this workbench in the workspace.\n> Name can only contain alphanumeric characters and '-' and can be atmost 25 characters long\n+sort=1\n+message=3 to 32 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    home_directory_size: conint(ge=5, le=64000) = Field(
        20,
        description="+label=Home Directory Size in GB (Persistent)\n+usage=Size of the home directory for the workbench (Persistent Storage)\n+sort=6",
    )
    resources: Optional[Resources] = None
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be set in the workbench pod(s)\n+sort=10110",
    )
    mounts: Optional[List[Union[SecretMount, StringDataMount, VolumeMount]]] = Field(
        None,
        description="+usage=Configure data to be mounted to workbench pod(s)\n+sort=10111",
    )
    service_account: Optional[str] = Field(
        None,
        description="+usage=Service account that this workload should use\n+sort=10113",
    )


class Build(BaseModel):
    """
    +docs=Describes how we build our code into a Docker image.
    +label=Source Code (Build and deploy source code)
    +icon=fa-code
    """

    type: constr(regex=r"^build$") = Field(..., description="+value=build")
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )
    build_source: Union[RemoteSource, GitSource, LocalSource] = Field(
        ...,
        description="+docs=Source code location.\n+label=Fetch source code to build and deploy\n+icon=fa-code\n+sort=1",
    )
    build_spec: Union[DockerFileBuild, PythonBuild] = Field(
        ...,
        description="+docs=Instructions to build a container image out of the build source\n+label=Build using DockerFile or using Buildpack\n+icon=fa-wrench\n+sort=2",
    )


class Canary(BaseModel):
    """
    +docs=This strategy brings up the new release without bringing the older release down. Traffic is shifted from the older release to the newer release in a staged manner.
    This can help with verifying the health of the new release without shifting complete traffic.
    +label=Canary strategy
    """

    type: constr(regex=r"^canary$") = Field(..., description="+value=canary")
    steps: List[CanaryStep] = Field(
        ...,
        description="+docs=These steps would be executed in order to enable shifting of traffic slowly from stable to canary version\n+label=Steps",
    )


class Codeserver(BaseWorkbenchInput):
    """
    +docs=Describes the configuration for the code server
    """

    type: constr(regex=r"^codeserver$") = Field(..., description="+value=Code Server")
    image: Union[CodeserverImage, CustomCodeserverImage] = Field(
        ..., description="+usage=Pick a codeserver image to deploy\n+sort=2"
    )
    auth: Optional[BasicAuthCreds] = None


class CoreNATSOutputConfig(BaseModel):
    """
    +docs=Describes the configuration for the output Core NATS worker
    +label=Core NATS Output Config
    """

    type: constr(regex=r"^core-nats$") = Field(..., description="+value=core-nats")
    nats_url: str = Field(
        ..., description="+label=NATS URL\n+usage=Output NATS URL\n+sort=1"
    )
    root_subject: constr(regex=r"^[a-zA-Z0-9][a-zA-Z0-9\-.]+[a-zA-Z0-9]$") = Field(
        ...,
        description="+label=Root Subject\n+usage=Root subject of output NATS\n+message=Output NATS root subject should only contain alphanumeric letters, dashes(-), and periods(.)\n+sort=2",
    )
    auth: Optional[NATSUserPasswordAuth] = None


class HealthProbe(BaseModel):
    """
    +docs=Describes the configuration for the Health Probe's
    To learn more you can go [here](https://docs.truefoundry.com/docs/add-health-checks-to-deployments)
    +icon=fa-heart
    """

    config: HttpProbe
    initial_delay_seconds: conint(ge=0, le=36000) = Field(
        0,
        description="+usage=Number of seconds after the container is started before the first probe is initiated.",
    )
    period_seconds: conint(ge=1, le=36000) = Field(
        10, description="+usage=How often, in seconds, to execute the probe."
    )
    timeout_seconds: conint(ge=1, le=36000) = Field(
        1, description="+usage=Number of seconds after which the probe times out."
    )
    success_threshold: conint(ge=1, le=100) = Field(
        1,
        description="+usage=Minimum consecutive successes for the probe to be considered successful after having failed.",
    )
    failure_threshold: conint(ge=1, le=100) = Field(
        3,
        description="+usage=Number of consecutive failures required to determine the container is not alive (liveness probe) or not ready (readiness probe).",
    )


class Helm(BaseModel):
    type: constr(regex=r"^helm$") = Field(..., description="+value=helm")
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,30}[a-z0-9]$") = Field(
        ...,
        description="+sort=1\n+message=3 to 32 lower case characters long alphanumeric word, may contain - in between, cannot start with a number\n+usage=Name of the Helm deployment. This will be set as the release name of the chart you are deploying.",
    )
    labels: Optional[Dict[str, str]] = Field(
        None, description="+label=Labels\n+usage=Add labels to base argo app"
    )
    source: Union[HelmRepo, OCIRepo] = Field(
        ..., description="+label=Source helm repository\n+sort=2"
    )
    values: Optional[Dict[str, Any]] = Field(
        None, description="+label=Values\n+usage=Values file as block file"
    )
    kustomize: Optional[Kustomize] = None


class Job(BaseModel):
    """
    +docs=Describes the configuration for the job
    """

    type: constr(regex=r"^job$") = Field(..., description="+value=job")
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,30}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the job\n+sort=1\n+message=3 to 32 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    image: Union[Build, Image] = Field(
        ...,
        description="+docs=Specify whether you want to deploy a Docker image or build and deploy from source code\n+label=Deploy a Docker image or build and deploy from source code\n+icon=fa-solid fa-cloud-arrow-up:#21B6A8\n+sort=200",
    )
    trigger: Union[Manual, Schedule] = Field(
        {"type": "manual"}, description="+docs=Specify the trigger\n+sort=300"
    )
    trigger_on_deploy: bool = Field(
        False,
        description="+docs=Trigger on deploy\n+sort=350\n+usage=Trigger the job after deploy immediately",
    )
    params: Optional[List[Param]] = Field(
        None,
        description="+label=Params for input\n+usage=Configure params and pass it to create different job runs\n+sort=400",
    )
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be injected in the service. [Docs](https://docs.truefoundry.com/docs/env-variables)\n+icon=fa-globe\n+sort=500",
    )
    resources: Optional[Resources] = None
    retries: conint(ge=0, le=10) = Field(
        0,
        description="+label=Retries\n+usage=Specify the maximum number of attempts to retry a job before it is marked as failed.\n+icon=fa-repeat\n+sort=700",
    )
    timeout: Optional[conint(le=432000, gt=0)] = Field(
        None,
        description="+label=Timeout\n+usage=Job timeout in seconds.\n+icon=fa-clock\n+sort=800",
    )
    concurrency_limit: Optional[PositiveInt] = Field(
        None,
        description="+label=Concurrency Limit\n+usage=Number of runs that can run concurrently\n+icon=fa-copy\n+sort=900",
    )
    service_account: Optional[str] = Field(
        None,
        description="+usage=Service account that this workload should use\n+sort=1000",
    )
    mounts: Optional[List[Union[SecretMount, StringDataMount, VolumeMount]]] = Field(
        None, description="+usage=Configure data to be mounted to job pod(s)"
    )
    labels: Optional[Dict[str, str]] = Field(
        None, description="+label=Labels\n+usage=Add labels to service metadata"
    )


class KafkaInputConfig(BaseModel):
    """
    +docs=Describes the configuration for the input Kafka worker
    +label=Kafka Input Config
    """

    type: constr(regex=r"^kafka$") = Field(..., description="+value=kafka")
    bootstrap_servers: str = Field(
        ...,
        description="+label=Bootstrap servers\n+usage='Kafka Bootstrap servers - Comma separated list of Kafka brokers \"hostname:port\" to connect to for bootstrap'\n+sort=1",
    )
    topic_name: str = Field(
        ...,
        description="+label=Topic Name\n+usage=Kafka topic to subscribe to\n+sort=2",
    )
    consumer_group: str = Field(
        ...,
        description="+label=Consumer Group Name\n+usage=The name of the consumer group to join for dynamic partition assignment\n+sort=3",
    )
    tls: bool = Field(
        True, description="+label=TLS\n+usage=TLS configuration for SASL authentication"
    )
    wait_time_seconds: conint(ge=1, le=300) = Field(
        10,
        description="+label=Wait Time Seconds\n+usage=Wait timeout for long polling.",
    )
    auth: Optional[KafkaSASLAuth] = None


class KafkaOutputConfig(BaseModel):
    """
    +docs=Describes the configuration for the output Kafka worker
    +label=Kafka Output Config
    """

    type: constr(regex=r"^kafka$") = Field(..., description="+value=kafka")
    bootstrap_servers: str = Field(
        ...,
        description="+label=Bootstrap servers\n+usage='Kafka Bootstrap servers - Comma separated list of Kafka brokers \"hostname:port\" to connect to for bootstrap'\n+sort=1",
    )
    topic_name: str = Field(
        ..., description="+label=Topic Name\n+usage=Kafka topic to publish to\n+sort=2"
    )
    tls: bool = Field(
        True, description="+label=TLS\n+usage=TLS configuration for SASL authentication"
    )
    auth: Optional[KafkaSASLAuth] = None


class NATSInputConfig(BaseModel):
    """
    +docs=Describes the configuration for the input NATS worker
    +label=NATS Input Config
    """

    type: constr(regex=r"^nats$") = Field(..., description="+value=nats")
    nats_url: str = Field(
        ..., description="+label=NATS URL\n+usage=Input NATS URL\n+sort=1"
    )
    stream_name: str = Field(
        ..., description="+label=Stream Name\n+usage=Name of the NATS stream\n+sort=2"
    )
    root_subject: constr(regex=r"^[a-zA-Z0-9][a-zA-Z0-9\-.]+[a-zA-Z0-9]$") = Field(
        ...,
        description="+label=Root Subject\n+usage=Root subject of input NATS\n+message=Input NATS root subject should only contain alphanumeric letters, dashes(-), and periods(.)\n+sort=3",
    )
    consumer_name: constr(regex=r"^[a-zA-Z0-9][a-zA-Z0-9\-_]+[a-zA-Z0-9]$") = Field(
        ...,
        description="+label=Consumer Name\n+usage=Consumer name of input NATS\n+message=Consumer name should only contain alphanumeric letters, dashes(-), and underscores(_)\n+sort=4",
    )
    wait_time_seconds: conint(ge=1, le=20) = Field(
        19,
        description="+label=Wait Time Seconds\n+usage=Wait timeout for long polling.\n+sort=5",
    )
    nats_metrics_url: Optional[constr(regex=r"^(http(s?)://).*$")] = Field(
        None,
        description="+label=NATS metrics URL\n+usage=URL for the NATS metrics endpoint. It is compulsory if you want to use NATS autoscaling.\n+message=NATS Metrics URL should be a valid HTTP/HTTPS URL",
    )
    auth: Optional[NATSUserPasswordAuth] = None


class NATSOutputConfig(BaseModel):
    """
    +docs=Describes the configuration for the output NATS worker
    +label=NATS Output Config
    """

    type: constr(regex=r"^nats$") = Field(..., description="+value=nats")
    nats_url: str = Field(
        ..., description="+label=NATS URL\n+usage=Output NATS URL\n+sort=1"
    )
    root_subject: constr(regex=r"^[a-zA-Z0-9][a-zA-Z0-9\-.]+[a-zA-Z0-9]$") = Field(
        ...,
        description="+label=Root Subject\n+usage=Root subject of output NATS\n+message=Output NATS root subject should only contain alphanumeric letters, dashes(-), and periods(.)\n+sort=2",
    )
    auth: Optional[NATSUserPasswordAuth] = None


class Notebook(BaseWorkbenchInput):
    """
    +docs=Describes the configuration for the service
    """

    type: constr(regex=r"^notebook$") = Field(..., description="+value=notebook")
    image: Union[
        TruefoundryImageBase,
        TruefoundryImageFull,
        CustomNotebookImage,
        TruefoundryImageCuda1180,
        TruefoundryImageCuda1211,
    ] = Field(
        ...,
        description="+usage=Changes made to the root directory `/` will not be persisted across notebook restarts\n+sort=2",
    )
    auth: Optional[BasicAuthCreds] = None
    cull_timeout: conint(ge=5) = Field(
        30,
        description="+label=Stop after (minutes of inactivity)\n+usage=Stop the notebook instance after this much time in minutes of inactivity.\nThe notebook instance will be stopped even if the notebook is open in your browser, but nothing is running on the notebook.\n+sort=5",
    )


class SSHServer(BaseWorkbenchInput):
    """
    +docs=Describes the configuration for the ssh server
    """

    type: constr(regex=r"^ssh-server$") = Field(..., description="+value=SSH Server")
    image: Union[SSHServerImage, CustomSSHServerImage] = Field(
        ..., description="+usage=Pick a ssh server image to deploy\n+sort=2"
    )
    ssh_public_key: str = Field(
        ...,
        description="+label: SSH Public Key\n+usage=Add Your SSH Public Key, this will be used to authenticate you to the SSH Server.\n+sort=4",
    )


class Volume(BaseModel):
    type: constr(regex=r"^volume$") = Field(..., description="+value=volume")
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,30}[a-z0-9]$") = Field(
        ...,
        description="+sort=1\n+message=3 to 32 lower case characters long alphanumeric word, may contain - in between, cannot start with a number\n+usage=Name of the Volume. This will be set as the volume name.",
    )
    storage_class: str = Field(..., description="+label=Storage Class\n+sort=2")
    size: conint(ge=1, le=64000) = Field(
        ..., description="+sort=3\n+label=Size\n+unit=Gi\n+usage=Size of volume in Gi"
    )
    volume_browser: Optional[VolumeBrowser] = None


class WorkerConfig(BaseModel):
    input_config: Union[SQSInputConfig, NATSInputConfig, KafkaInputConfig] = Field(
        ..., description="+label=Input Config\n+usage=Input Config\n+sort=1"
    )
    output_config: Optional[
        Union[
            SQSOutputConfig, NATSOutputConfig, CoreNATSOutputConfig, KafkaOutputConfig
        ]
    ] = Field(None, description="+label=Output Config\n+usage=Output Config\n+sort=2")
    num_concurrent_workers: conint(ge=1, le=10) = Field(
        1,
        description="+label=Number of Concurrent Workers\n+usage=Number of concurrent workers to spawn for the processor\n+sort=3",
    )


class BaseService(BaseModel):
    name: constr(regex=r"^[a-z][a-z0-9\-]{1,30}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the service. This uniquely identifies this service in the workspace.\n> Name can only contain alphanumeric characters and '-' and can be atmost 25 characters long\n+sort=1\n+message=3 to 32 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    image: Union[Build, Image] = Field(
        ...,
        description="+docs=Specify whether you want to deploy a Docker image or build and deploy from source code\n+label=Deploy a Docker image or build and deploy from source code\n+icon=fa-solid fa-cloud-arrow-up:#21B6A8\n+sort=2",
    )
    artifacts_download: Optional[ArtifactsDownload] = None
    resources: Optional[Resources] = None
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be injected in the service. [Docs](https://docs.truefoundry.com/docs/env-variables)\n+icon=fa-globe\n+sort=6",
    )
    ports: List[Port] = Field(
        ...,
        description="+docs=Specify the ports you want the service to be exposed to\n+label=Ports (Ports to route customer traffic to)\n+icon=fa-plug\n+sort=4",
    )
    service_account: Optional[str] = Field(
        None, description="+usage=Service account that this workload should use"
    )
    mounts: Optional[List[Union[SecretMount, StringDataMount, VolumeMount]]] = Field(
        None,
        description="+usage=Configure data to be mounted to service pod(s)\n+sort=10011",
    )
    labels: Optional[Dict[str, str]] = Field(
        None, description="+label=Labels\n+usage=Add labels to service metadata"
    )
    kustomize: Optional[Kustomize] = None
    liveness_probe: Optional[HealthProbe] = None
    readiness_probe: Optional[HealthProbe] = None


class Service(BaseService):
    """
    +docs=Describes the configuration for the service
    """

    type: constr(regex=r"^service$") = Field(..., description="+value=service")
    replicas: Union[confloat(ge=0.0, le=500.0), ServiceAutoscaling] = Field(
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone\n+sort=4",
    )
    allow_interception: bool = Field(
        False,
        description="+label=Allow intercepts\n+usage=Whether to allow intercepts to be applied for this service.\nThis would inject an additional sidecar in each pod of the service. Not recommended on production",
    )
    rollout_strategy: Optional[Union[Rolling, Canary, BlueGreen]] = Field(
        None,
        description="+label=Rollout strategy\n+usage=Strategy to dictate how a rollout should happen when a new release for this service is made",
    )


class AsyncService(BaseService):
    """
    +docs=Describes the configuration for the async-service
    """

    type: constr(regex=r"^async-service$") = Field(
        ..., description="+value=async-service"
    )
    replicas: Union[confloat(ge=0.0, le=500.0), AsyncServiceAutoscaling] = Field(
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone",
    )
    rollout_strategy: Optional[Rolling] = None
    worker_config: WorkerConfig
    sidecar: Optional[AsyncProcessorSidecar] = None


class Application(BaseModel):
    __root__: Union[
        Service, AsyncService, Job, Notebook, Codeserver, SSHServer, Helm, Volume
    ]
